PROGRAM AUTOCOR(INPUT,OUTPUT,ENTREEC,SORTIEC,ENTREEB,CLASSEF,
                             LIEN,MATBIT);
LABEL 777;
CONST   MAXOBJ=500;             (* Maximum number of objects       *)
        MAXTRIANGLES=3000;      (* N. triangles in triangulation   *)
        NBBITS=59;              (* N. bits in a Pascal SET         *)
        NBBITSL1=58;            (* NBBITS - 1                      *)
        MAXMOTS=8;              (* MAXOBJ DIV NBBITS               *)
        MAXLONG=1000;
        MAXCLASSES=50;          (* Maximum number of classes       *)
        EPSILON=10E-7;
        MAXREEL=1.70E38;        (* IBM: 10E74 ; CDC: 10E320        *)
        NBMOTSSIMIL=1;          (* Parameter NBMOTS from SIMIL     *)
        NBCAR=10;               (* Parameter NBCAR from SIMIL      *)
        PROBPREC=0.001;         (* Precision for exact probability *)
        LANG=1;                 (* Parametre de langue: 1= FRANCAIS
                                   Language parameter : 2= ENGLISH *)
(* =============================================================== *)
(* Program for computing spatial correlograms, for quantitative    *)
(* or nominal data.                                                *)
(* Files used in this program:                                     *)
(* INPUT and OUTPUT are connected to the user's terminal.          *)
(* ENTREEC: Raw input data file.                                   *)
(*   For options 12 and 13, this file must contain 3 columns:      *)
(*        (X coordinate)  (Y coordinate)  (variable)               *)
(*   For all the other options, it contains only the values        *)
(*   of the variable (either quantitative measures, or nominal     *)
(*   variable classes).                                            *)
(*   For option 13 (Delaunay), if the user wishes to give the      *)
(*   program his own list of constraint lines, each of these lines *)
(*   must be provided to the program as one additional line at the *)
(*   end of the list of objects. Each line contains the coordinates*)
(*   of two points (that mark the ends of the constraint line),    *)
(*   which means 4 coordinates.                                    *)
(* SORTIEC: Computation results (correlogram, etc.)                *)
(* ENTREEB: a binary DISTANCE matrix produced by SIMIL, for        *)
(*   option 0 of the program.                                      *)
(* CLASSEF: a TEXT file containing an upper triangular matrix or   *)
(*   a square matrix of distance classes, to be used by option 14  *)
(*   of this program. If this program is asked to write an upper   *)
(*   triangular matrix of distance classes (to be used by program  *)
(*   MANTEL3 to compute a Mantel correlogram), that matrix is also *)
(*   written onto file CLASSEF.                                    *)
(* LIEN: a TEXT file containing the list of link edges. This file  *)
(*   can be used in input (for option 15) or produced as an output,*)
(*   in which case it can be used by programs BIOGEO, KMEANS and   *)
(*   PRECEL.                                                       *)
(* MATBIT: a TEXT file containing information useful for debugging.*)
(*   Option disabled in this version of the program.               *)
(* =============================================================== *)
TYPE    ENSBITS=0..NBBITSL1;
        ENS=SET OF ENSBITS;
        VECBITS=ARRAY[0..MAXMOTS] OF ENS;
        MATBITS=ARRAY[0..MAXOBJ] OF VECBITS;
        MATREAL=ARRAY[0..MAXOBJ] OF REAL;
        MATLONG=ARRAY[1..MAXLONG] OF REAL;
        PTRMATLONG=@MATLONG;
        MATPREAL=ARRAY[0..MAXCLASSES] OF REAL;
        MATFREQ=ARRAY[0..MAXCLASSES]OF INTEGER;
        PTRFREQ=@MATFREQ;
        PTRMATPREAL=@MATPREAL;
        DATA=RECORD CASE INTEGER OF
                2:(RE:REAL);
                1:(INT:INTEGER);
                3:(MATCAR:PACKED ARRAY[1..NBCAR] OF CHAR)
                END;
        COORDC=RECORD OBJ1,OBJ2,OBJ3:INTEGER;END;
        CIRCLE=ARRAY[0..MAXTRIANGLES] OF COORDC;
        PTRCIRCLE=@CIRCLE;
        FICHINT=FILE OF INTEGER;
VAR     NOBJ,NIVEAU,NBMOTS,MOUVEMENT,HAUTEUR,LARGEUR,INDICE,MAX,NBCR,
        NBLIENS,
        TOTAL,OPTCLASSES,NBCLASSES,ASAUTER:INTEGER;
        MOYENNE,VARIANCE,B2,S1,S2,W,IMORAN,CGEARY,KALPHA,TRENTEPOURCENT,
        NOBJREAL,PROBI,PROBC,NRNS,NR2NS2,NRNSNT,NRNSNTNU,N2,N3,N4:REAL;
        DIST1,DISTN,DEJAFAIT:MATBITS;
        CERCLE:PTRCIRCLE;
        DONNEES:PTRMATLONG;
        PTRREAL:PTRMATPREAL;
        FREQUENCES:PTRFREQ;
        MAT:MATFREQ;
        NOCLASSE:ARRAY[1..MAXOBJ] OF INTEGER;
        TEMP:VECBITS;
        COORDV,COORDH,MESURE:MATREAL;
        ELIMINATION,NOMINAL,SYMMETRIC,PGTRENTEPOURCENT,
        LIENSECRITS:BOOLEAN;
        ENTREEC,LIEN,MATBIT,CLASSEF,SORTIEC:TEXT;
        FICHTEMP:FICHINT;
        ENTREEB:FILE OF DATA;
 
FUNCTION GAUSS(X:REAL):REAL;
VAR Y,Z,W:REAL;
BEGIN
                (* REFERENCE: CACM                       *)
                (* ALGORITHME: 209                       *)
                (* AUTEUR:D.IBBETSON                     *)
                (* TITRE:GAUSS                           *)
                (* TRADUIT EN PASCAL PAR A.VAUDOR        *)
IF X=0 THEN Z:=0
ELSE
        BEGIN
        Y:=ABS(X)/2;
        IF Y>=3 THEN Z:=1
        ELSE IF Y<1 THEN
                BEGIN
                W:=Y*Y;
                Z:=((((((((0.000124818987*W
                -0.001075204047)*W+0.005198775019)*W
                -0.019198292004)*W+0.059054035642)*W
                -0.151968751364)*W+0.319152932694)*W
                -0.531923007300)*W+0.797884560593)*Y*2
                END
           ELSE
                BEGIN
                Y:=Y-2;
                Z:=(((((((((((((-0.000045255659*Y
                +0.000152529290)*Y-0.000019538132)*Y
                -0.000676904986)*Y+0.001390604284)*Y
                -0.000794620820)*Y-0.002034254874)*Y
                +0.006549791214)*Y-0.010557625006)*Y
                +0.011630447319)*Y-0.009279453341)*Y
                +0.005353579108)*Y-0.002141268741)*Y
                +0.000535310849)*Y+0.999936657524
                END;
        END;
IF X>0 THEN GAUSS:=(Z+1)/2 ELSE GAUSS:=(1-Z)/2;
END;                (* GAUSS *)
 
FUNCTION NORMDEV(P:REAL):REAL;
VAR S,X,Z:REAL;
BEGIN
IF P>=1.0 THEN NORMDEV:=MAXREEL
ELSE IF P <= 0 THEN NORMDEV:=-MAXREEL
   ELSE
        BEGIN
        IF P>=0.5 THEN X:=1-P ELSE X:=P;
        S:=SQRT(-2*LN(X));
        X:=((-7.49101*S-448.047)*S-1266.846)/
                (((S+109.8371)*S+748.189)*S+498.003)+S;
        IF P < 0.5 THEN X:=-X;
        Z:=P-GAUSS(X);
        S:=EXP(-X*X/2)/2.506628275;
        Z:=Z/S;S:=X*X;
        NORMDEV:=(((((((102*S+326)*S+127)*X*Z/6
                +(24*S+46)*S+7)*Z/40+(0.75*S
                +0.875)*X)*Z
                +S+0.5)*Z/3.0+X*0.5)*Z+1)*Z+X;
        END;
END;                (*NORMDEV *)
 
FUNCTION CHIPROB(X:REAL;F:INTEGER):REAL;
VAR A,Y,S,E,C,Z:REAL;
    EVEN:BOOLEAN;
BEGIN
IF (X<0)OR (F<1) THEN CHIPROB:=0
ELSE
   IF X=0 THEN CHIPROB:=1
   ELSE
        BEGIN
        A:=0.5*X;EVEN:=2*(F DIV 2)=F;
        IF (EVEN) OR(F>2) THEN Y:=EXP(-A);
        IF EVEN THEN S:=Y ELSE S:=2*(GAUSS(-SQRT(X)));
        IF F>2 THEN
                BEGIN
                X:=0.5*(F-1);
                IF EVEN THEN Z:=1 ELSE Z:=0.5;
                IF EVEN THEN E:=1 ELSE E:=0.564189583548/SQRT(A);
                C:=0;
                WHILE Z<X DO
                        BEGIN
                        E:=E*A/Z;
                        C:=C+E;
                        Z:=Z+1;
                        END;
                CHIPROB:=C*Y+S;
                END
        ELSE CHIPROB:=S;
        END;
END;        (* CHIPROB *)
 
FUNCTION VALIDEBOOL:BOOLEAN;
VAR RESPONSE:CHAR;
BEGIN
READ(INPUT,RESPONSE);
WHILE NOT(RESPONSE IN['0','O','o','Y','y','N','n'])
DO READ(INPUT,RESPONSE);
VALIDEBOOL:=RESPONSE IN['0','O','o','Y','y'];
END;
 
FUNCTION MEMBRE(IND:INTEGER;VAR MM:VECBITS):BOOLEAN;
VAR I,J:INTEGER;
BEGIN
I:=IND DIV NBBITS;
J:=IND MOD NBBITS;
MEMBRE:=J IN MM[I];
END;
 
PROCEDURE ERREUR(I,J:INTEGER);
BEGIN
IF LANG=1 THEN BEGIN
CASE I OF
        1:WRITELN(' Les points superposes ne sont pas permis ',
               'pour cette fonction: ligne ',J:3);
        4:WRITELN(' Il y a une distance trop grande dans l''une ',
           'des classes de distance');
        5:WRITELN(' Numero d''objet dans le fichier LIENS > NOBJ');
        6:WRITELN(' Numero d''objet dans le fichier LIENS <= 0');
        7:WRITELN(' Numero de classe <= 0 dans le fichier de ',
           'donnees, objet ',J:3);
        8:WRITELN(' Numero de classe > n. maximum ',
          'permis: fichier de donnees, objet ',J:3);
        9:WRITELN(' Trop de triangles: augmenter le parametre ',
          'MAXTRIANGLES du programme');
      10:WRITELN(' Valeur incalculable car la variance = 0');
        END;
END ELSE BEGIN
CASE I OF
        1:WRITELN(' Overlying points are not allowed for this '
               ,'function: line ',J:3);
        4:WRITELN(' There is too large a distance in one of the ',
          'distance classes');
        5:WRITELN(' No. of an object in file LINKS > NOBJ');
        6:WRITELN(' No. of an object in file LINKS <= 0');
        7:WRITELN(' No. of a class <= 0 in the data file, object ',
           J:3);
        8:WRITELN(' No. of a class > max. number allowed: ',
          'Data file, object ',J:3);
        9:WRITELN(' Too many triangles: increase parameter ',
          'MAXTRIANGLES of the program');
        10:WRITELN(' Value cannot be computed because variance = 0');
        END;
END;
GOTO 777;
END;
 
PROCEDURE POSITIONNE;
VAR I:INTEGER;
BEGIN
RESET(ENTREEB);
FOR I:=1 TO ASAUTER DO GET(ENTREEB);
END;
 
PROCEDURE VALIDE(VAR J:INTEGER;LIMINF,LIMSUP:INTEGER);
VAR TERMINE:BOOLEAN;
BEGIN
READLN(INPUT);
TERMINE:=FALSE;
REPEAT
        READ(INPUT,J);
        IF (J >= LIMINF) AND (J <= LIMSUP) THEN TERMINE:=TRUE
        ELSE BEGIN
           IF LANG=1 THEN
           WRITELN('   S.V.P. ',LIMINF:3,'...',LIMSUP:3,'   ')
           ELSE
           WRITELN('   PLEASE ',LIMINF:3,'...',LIMSUP:3,'   ');
           END;
UNTIL TERMINE;
END;                (* VALIDE *)
 
PROCEDURE NULL(VAR M:MATBITS);
VAR I,J:INTEGER;
BEGIN
FOR I:=1 TO NOBJ DO
   FOR J:=0 TO NBMOTS DO M[I,J]:=[];
END;
 
PROCEDURE NULLVEC(VAR M:VECBITS);
VAR I:INTEGER;
BEGIN
FOR I:=0 TO NBMOTS DO M[I]:=[];
END;
 
FUNCTION CARD(VAR A:ENS):INTEGER;
VAR I,C:INTEGER;
BEGIN
C:=0;
IF A<>[] THEN
      BEGIN
      FOR I:=0 TO NBBITSL1 DO
          IF I IN A THEN C:=C+1;
      END;
CARD:=C;
END;
 
PROCEDURE UNIONMAT(VAR M1,M2:MATBITS;VAR C:INTEGER);
VAR I,J:INTEGER;
    TEMP:ENS;
BEGIN
C:=0;
FOR I:=1 TO NOBJ DO
        BEGIN
        FOR J:=0 TO NBMOTS DO
                BEGIN
                TEMP:=M1[I,J]+M2[I,J];
                C:=C+CARD(TEMP);
                M1[I,J]:=TEMP;
                END;
        END;
END;
 
PROCEDURE UNION(VAR M1,M2:VECBITS);
VAR I:INTEGER;
BEGIN
FOR I:=0 TO NBMOTS DO M1[I]:=M1[I]+M2[I];
END;
 
PROCEDURE ADD(IND1,IND2:INTEGER;VAR M:MATBITS);
VAR I:INTEGER;
BEGIN
I:=IND2 DIV NBBITS;
M[IND1,I]:=M[IND1,I]+[IND2 MOD NBBITS];
I:=IND1 DIV NBBITS;
M[IND2,I]:=M[IND2,I]+[IND1 MOD NBBITS];
END;
 
PROCEDURE SUB(IND1,IND2:INTEGER;VAR M:MATBITS);
VAR I:INTEGER;
BEGIN
I:=IND2 DIV NBBITS;
M[IND1,I]:=M[IND1,I]-[IND2 MOD NBBITS];
I:=IND1 DIV NBBITS;
M[IND2,I]:=M[IND2,I]-[IND1 MOD NBBITS];
END;
 
PROCEDURE AJOUTE(IND:INTEGER;VAR M:VECBITS);
VAR I:INTEGER;
BEGIN
I:=IND DIV NBBITS;
M[I]:=M[I]+[IND MOD NBBITS];
END;
 
PROCEDURE ENLEVE(IND:INTEGER;VAR M:VECBITS);
VAR I:INTEGER;
BEGIN
I:=IND DIV NBBITS;
M[I]:=M[I]-[IND MOD NBBITS];
END;
 
PROCEDURE DIFFERENCE(VAR M1,M2:VECBITS);
VAR I:INTEGER;
BEGIN
FOR I:=0 TO NBMOTS DO M1[I]:=M1[I]-M2[I];
END;
 
PROCEDURE LISLIENS;
VAR I,J:INTEGER;
BEGIN
NULL(DIST1);
RESET(LIEN);
WHILE NOT(EOF(LIEN)) DO
        BEGIN
        READ(LIEN,I,J);
        IF (I > NOBJ)OR(J > NOBJ) THEN ERREUR(5,0);
        IF (I < 1 )OR(J < 1   )   THEN ERREUR(6,0);
        AJOUTE(J,DIST1[I]);
        WHILE NOT((EOF(LIEN))OR(LIEN@ IN['0'..'9'])) DO GET(LIEN);
        END;
END;   (* LISLIEN *)
 
PROCEDURE PARAMETRES;
VAR J:INTEGER;
BEGIN
IF LANG=1 THEN BEGIN
WRITELN(' Votre fichier de donnees est-il deja en classes ?  ');
WRITELN(' Autrement dit, desirez-vous analyser',
        ' des DONNEES QUALITATIVES ? ');
END ELSE BEGIN
WRITELN(' Is your data file already in classes ?  ');
WRITELN(' In other words, are you analyzing NOMINAL DATA ?');
END;
NOMINAL:=VALIDEBOOL;
WRITELN(' OPTIONS: ');
IF LANG=1 THEN BEGIN
WRITELN('   0: Matrice de distances de SIMIL (Fichier "ENTREEB")');
WRITELN;
WRITELN('  MOVEMENTS DANS UNE SEULE DIRECTION:');
WRITELN('   1: Mouvement horizontal (Lignes)');
WRITELN('   2: Mouvement vertical (Colonnes)');
WRITELN('   3: Mouvement diagonal (pente positive)');
WRITELN('   4: Mouvement diagonal (pente negative)');
WRITELN;
WRITELN('  JEU D''ECHECS, MOUVEMENTS DIRECT SEULEMENT:');
WRITELN('   5: Mouvement de la tour');
WRITELN('   6: Mouvement du fou');
WRITELN('   7: Mouvement de la reine');
WRITELN;
WRITELN('  JEU D''ECHECS, MOUVEMENTS DIRECTS ET INDIRECTS:');
WRITELN('   8: Mouvement de la tour');
WRITELN('   9: Mouvement du fou');
WRITELN('  10: Mouvement de la reine');
WRITELN;
WRITELN('  11: Distance euclidienne, points en grille reguliere');
WRITELN;
WRITELN('  POINTS DISPOSES DE FACON IRREGULIERE:');
WRITELN('  12: Graphique de Gabriel');
WRITELN('  13: Triangulation de Delaunay');
WRITELN('  14: Votre propre matrice de classes de distance');
WRITELN('  15: Votre propre liste de liens (attacher fichier "LIENS")');
END ELSE BEGIN
WRITELN('   0: Matrix of distances from SIMIL (File "ENTREEB")');
WRITELN;
WRITELN('  MOVEMENTS IN ONE DIRECTION ONLY:');
WRITELN('   1: Horizontal movement (Rows)');
WRITELN('   2: Vertical movement (Columns)');
WRITELN('   3: Diagonal movement (positive slope)');
WRITELN('   4: Diagonal movement (negative slope)');
WRITELN;
WRITELN('  DIRECT CHESS MOVEMENTS ONLY:');
WRITELN('   5: Rook''s move');
WRITELN('   6: Bishop''s move');
WRITELN('   7: Queen''s move');
WRITELN;
WRITELN('  DIRECT AND INDIRECT CHESS MOVEMENTS:');
WRITELN('   8: Rook''s move');
WRITELN('   9: Bishop''s move');
WRITELN('  10: Queen''s move');
WRITELN;
WRITELN('  11: Euclidean distance on regular grid');
WRITELN;
WRITELN('  NON-REGULAR POINT LOCATIONS:');
WRITELN('  12: Gabriel graph');
WRITELN('  13: Delaunay triangulation');
WRITELN('  14: Your own matrix of distance classes');
WRITELN('  15: Your own list of links (attach file "LINKS")');
END;
WRITELN('   ');
VALIDE(MOUVEMENT,0,15);
CASE MOUVEMENT OF
        0:BEGIN        END;
        1,2,3,4,5,6,7,8,9,10,11:BEGIN
                IF LANG=1 THEN
                WRITELN(' Largeur et hauteur de la grille reguliere?')
                ELSE
                WRITELN(' Width and height of the regular lattice?  ');
                READLN(INPUT);
                READ(INPUT,LARGEUR,HAUTEUR);
                NOBJ:=LARGEUR*HAUTEUR;
                NBMOTS:=NOBJ DIV NBBITS;
                END;
        12,13,14:BEGIN
                IF LANG=1 THEN
                WRITELN(' Nombre total de points ?   ')
                ELSE
                WRITELN(' Total number of points ?   ');
                VALIDE(NOBJ,0,MAXLONG);
                NBMOTS:=NOBJ DIV NBBITS;
                LARGEUR:=NOBJ;HAUTEUR:=1;
                IF MOUVEMENT IN[12,13] THEN
                  BEGIN
                  IF MOUVEMENT=13 THEN
                     BEGIN
                     IF LANG=1 THEN
                     WRITELN(' Nombre de points de contrainte? (-1 = ',
                       'contrainte rectangulaire)')
                     ELSE
                     WRITELN(' Number of constraint points? (-1 = ',
                       'rectangular constraint)');
                     VALIDE(NBCR,-1,10000);
                     IF LANG=1 THEN
                     WRITELN(' Elimination des traits qui se coupent?')
                     ELSE
                     WRITELN(' Elimination of crossing edges?  ');
                                ELIMINATION:=VALIDEBOOL;
                                END;
                        END;
                END;
        15:        BEGIN
                IF LANG=1 THEN
                WRITELN(' Nombre total de points ?   ')
                ELSE
                WRITELN(' Total number of points ?   ');
                VALIDE(NOBJ,0,MAXLONG);
                NBMOTS:=NOBJ DIV NBBITS;
                LARGEUR:=NOBJ;HAUTEUR:=1;
                LISLIENS;
                END;
        END;
WRITELN;
END;
 
PROCEDURE INIT;
VAR I:INTEGER;
 
        PROCEDURE ECR(VAR SORTIEC:TEXT);
        BEGIN
        WRITELN(SORTIEC);
        IF LANG=1 THEN BEGIN
        WRITELN(SORTIEC,' A U T O C O R R E L A T I O N   ',
          'S P A T I A L E   ');
        WRITELN(SORTIEC);
        WRITELN(SORTIEC,' pour donnees quantitatives ou qualitatives.');
        WRITELN(SORTIEC,' Version IBM 2.0B');
        WRITELN(SORTIEC,' Auteur: Alain VAUDOR');
        END ELSE BEGIN
        WRITELN(SORTIEC,' S P A T I A L   ',
          'A U T O C O R R E L A T I O N   ');
        WRITELN(SORTIEC);
        WRITELN(SORTIEC,' for quantitative or nominal data.');
        WRITELN(SORTIEC,' Version IBM 2.0B');
        WRITELN(SORTIEC,' Author: Alain VAUDOR');
        END;
        WRITELN(SORTIEC);
        END;
BEGIN
SYMMETRIC:=TRUE;
TOTAL:=0;
ECR(OUTPUT);
RESET(ENTREEC);
REWRITE(SORTIEC);
ECR(SORTIEC);
END;
 
PROCEDURE ENTETE(VAR ASAUTER:INTEGER);
VAR I,J:INTEGER;
    DONNEE:DATA;
BEGIN
RESET(ENTREEB);
DONNEE:=ENTREEB@;
NOBJ:=DONNEE.INT;
GET(ENTREEB);GET(ENTREEB);
FOR I:=1 TO 13 DO
   FOR J:=1 TO NBMOTSSIMIL DO GET(ENTREEB);
DONNEE:=ENTREEB@;
GET(ENTREEB);
ASAUTER:=3+13*NBMOTSSIMIL;
IF DONNEE.INT<>0 THEN
        BEGIN
        FOR I:=1 TO NOBJ DO
           FOR J:=1 TO NBMOTSSIMIL DO GET(ENTREEB);
        ASAUTER:=ASAUTER+NOBJ*NBMOTSSIMIL;
        END;
END;                        (* ENTETE *)
 
PROCEDURE ECRITCLASSES;
VAR I,J,K:INTEGER;
BEGIN
IF LANG=1 THEN BEGIN
WRITELN(' Desirez-vous faire ecrire la matrice CLASSEF des classes');
WRITELN(' de distance, pour le correlogramme de Mantel ?  ');
END ELSE BEGIN
WRITELN(' Do you wish to have the CLASSEF file of distance');
WRITELN(' classes written out, for Mantel correlogram ?  ');
END;
IF VALIDEBOOL THEN
        BEGIN
        REWRITE(CLASSEF);
        RESET(FICHTEMP);
        FOR I:=1 TO NOBJ DO
                BEGIN
                WRITE(CLASSEF,' ');
                FOR J:=1 TO I DO WRITE(CLASSEF,' ':3);
                FOR J:=I+1 TO NOBJ DO
                        BEGIN
                        K:=FICHTEMP@;
                        GET(FICHTEMP);
                        WRITE(CLASSEF,K:3);
                        IF J MOD 20=0 THEN WRITELN(CLASSEF);
                        END;
                IF NOBJ MOD 20 <> 0 THEN WRITELN(CLASSEF);
                END;
        END;
END;
 
PROCEDURE CLASSE(VAR DONNEES:MATLONG;VAR MATCLASSES:MATPREAL;VAR
     FREQUENCES:MATFREQ);
LABEL 77;
VAR I,J,K,GENRE,CLASSEFAITE,NOBJ1,LENGTH,RESTANT,NC,OFFSET,
    LONGUEUR:INTEGER;
    DONNEE:DATA;
    MAX,MAXL,X,XREF,ETENDUE,XMAX:REAL;
 
        FUNCTION FOUILLEBINAIRE(X:REAL):INTEGER;
        VAR I,M2:INTEGER;
        BEGIN
        I:=(NBCLASSES+1) DIV 2;
        M2:=I;
        WHILE NOT(( X > MATCLASSES[I-1])AND(X<=MATCLASSES[I])) DO
                BEGIN
                M2:=(M2 +1)DIV 2;
                IF X <MATCLASSES[I] THEN I:=I-M2
                ELSE I:=I+M2;
                END;
        FOUILLEBINAIRE:=I;
        END;                (* FIN FOUILLE BINAIRE *)
 
        PROCEDURE SHELLSORT(VAR A: MATLONG;N: INTEGER);
        LABEL 1;
        VAR I,J,K,L,M,O: INTEGER;
            W:REAL;
        BEGIN
        M:=2*N-1 DIV 2;
        WHILE M<>0 DO
                BEGIN
                K:=N-M;
                FOR J:=1 TO K DO
                        BEGIN
                        I:=J;
                        WHILE I>=1 DO
                                BEGIN
                                IF A[I+M]>=A[I] THEN GOTO 1;
                                O:=I+M;
                                W:=A[I];A[I]:=A[O];A[O]:=W;
                                I:=I-M;
                                END;
        1:                END;
                M:=M  DIV 2;
                END;
        END;                (* FIN DE SHELLSORT *)
 
        PROCEDURE HISTO(ECR:BOOLEAN);
        VAR I,J:INTEGER;
        BEGIN
        POSITIONNE;
        FOR I:=1 TO NBCLASSES DO FREQUENCES[I]:=0;
        FOR I:=1 TO NOBJ1 DO
                BEGIN
                X:=ENTREEB@.RE;GET(ENTREEB);
                J:=TRUNC(X/ETENDUE)+1;
                FREQUENCES[J]:=FREQUENCES[J]+1;
                END;
        IF ECR THEN
           BEGIN
           MAX:=0;
           FOR I:=1 TO NBCLASSES DO
                   IF FREQUENCES[I] > MAX THEN MAX:=FREQUENCES[I];
           IF MAX > 60 THEN MAX:=60/MAX
           ELSE MAX:=1;
           FOR I:=1 TO NBCLASSES DO
                   BEGIN
                   WRITE('!');WRITE(SORTIEC,'!');
                   FOR J:=1 TO ROUND(FREQUENCES[I]*MAX) DO
                           BEGIN
                           WRITE('*');
                           WRITE(SORTIEC,'*');
                           END;
                   WRITELN;WRITELN(SORTIEC);
                   END;
           FOR I:=1 TO 60 DO
                   BEGIN
                   WRITE('_');
                   WRITE(SORTIEC,'_');
                   END;
           WRITELN;WRITELN(SORTIEC);
           END;
        END;                (* HISTO *)
BEGIN                (* CLASSE *)
NOBJ1:=NOBJ*(NOBJ-1) DIV 2;
MAXL:=-MAXREEL;
POSITIONNE;
FOR I:=1 TO NOBJ1 DO
        BEGIN
        X:=ENTREEB@.RE;GET(ENTREEB);
        IF X > MAXL THEN MAXL:=X;
        END;
REPEAT
     77:IF LANG=1 THEN
        WRITELN(' Classes equidistantes (0) ou equifrequentes (1)? ')
        ELSE
        WRITELN(' Equidistant (0) or equal frequency (1) classes?  ');
        VALIDE(GENRE,0,1);
        IF LANG=1 THEN BEGIN
          IF GENRE=0 THEN
          WRITELN(SORTIEC,' Classes equidistantes')
          ELSE WRITELN(SORTIEC,'Classes equifrequentes');
          WRITELN(' Nombre de classes   ');
        END ELSE BEGIN
          IF GENRE=0 THEN
          WRITELN(SORTIEC,' Equidistant classes')
          ELSE WRITELN(SORTIEC,'Equal frequency classes');
          WRITELN(' Number of classes   ');
        END;
        IF GENRE=0 THEN
                BEGIN
                VALIDE(NBCLASSES,0,MAXCLASSES);
                ETENDUE:=MAXL/NBCLASSES+EPSILON;
                X:=ETENDUE;
                FOR I:=1 TO NBCLASSES DO
                        BEGIN
                        MATCLASSES[I]:=X;
                        X:=X+ETENDUE;
                        END;
                IF LANG=1 THEN
                WRITELN(' Desirez-vous voir l''histogramme ? ')
                ELSE
                WRITELN(' Do you want to see the histogram ? ');
                HISTO(VALIDEBOOL);
                END
        ELSE
                BEGIN
                VALIDE(NBCLASSES,(NOBJ1+NOBJ1)DIV MAXLONG,MAXCLASSES);
                CLASSEFAITE:=0;
                XREF:=-MAXREEL;
                RESTANT:=NOBJ1;
                NC:=NBCLASSES;
                OFFSET:=0;
                REPEAT
                        XMAX:=MAXREEL;
                        LENGTH:=RESTANT DIV NC;
                        IF RESTANT MOD NC > 0 THEN LENGTH:=LENGTH+1;
                        LONGUEUR:=LENGTH-OFFSET;
                        IF LONGUEUR < 1 THEN
                        BEGIN
                        IF LANG=1 THEN
                        WRITELN(' Je ne puis creer autant de classes')
                        ELSE
                        WRITELN(' I cannot create that many classes');
                                GOTO 77;
                                END;
                        NC:=NC-1;
                        CLASSEFAITE:=CLASSEFAITE+1;
                        POSITIONNE;
                        J:=0;
                        FOR I:=1 TO NOBJ1 DO
                                BEGIN
                                X:=ENTREEB@.RE;GET(ENTREEB);
                                IF (X > XREF)AND(X <=XMAX) THEN
                                        BEGIN
                                        J:=J+1;
                                        DONNEES[J]:=X;
                                        END;
                                IF (J >= MAXLONG) OR(I=NOBJ1) THEN
                                        BEGIN
                                        SHELLSORT(DONNEES,J);
                                        XMAX:=DONNEES[LONGUEUR];
                                        IF CLASSEFAITE<> NBCLASSES
                                        THEN J:=LONGUEUR
                                        ELSE LONGUEUR:=J;
                                        END;
                                END;
                        RESTANT:=RESTANT-LENGTH;
                        XREF:=DONNEES[LONGUEUR];
                        OFFSET:=0;
                        WHILE DONNEES[LONGUEUR+OFFSET+1]=
                        DONNEES[LONGUEUR+OFFSET] DO
                                OFFSET:=OFFSET+1;
                        MATCLASSES[CLASSEFAITE]:=XREF;
                        FREQUENCES[CLASSEFAITE]:=LONGUEUR+OFFSET;
                UNTIL CLASSEFAITE = NBCLASSES;
                END;
        FOR I:=2 TO NBCLASSES DO
                IF MATCLASSES[I]=MATCLASSES[I-1] THEN ERREUR(4,0);
        IF LANG=1 THEN
        WRITELN('   CLASSE    LIMITE SUP.    FREQ.')
        ELSE
        WRITELN('   CLASS     UPPER LIMIT    FREQ.');
        FOR I:=1 TO NBCLASSES DO
             WRITELN(I:10,MATCLASSES[I]:10:5,FREQUENCES[I]:10);
        IF LANG=1 THEN WRITELN
        (' Prefereriez-vous un nombre/type different de classes?')
        ELSE
        WRITELN(' Would you prefer a different number/type of classes?')
UNTIL NOT(VALIDEBOOL);
REWRITE(FICHTEMP);
POSITIONNE;
IF GENRE=0 THEN
        BEGIN
        FOR I:=1 TO NOBJ1 DO
                BEGIN
                X:=ENTREEB@.RE;GET(ENTREEB);
                J:=TRUNC(X/ETENDUE)+1;
                FICHTEMP@:=J;
                PUT(FICHTEMP);
                END;
        END
ELSE
        BEGIN
        MATCLASSES[0]:=-MAXREEL;
        FOR I:=1 TO NOBJ1 DO
                BEGIN
                X:=ENTREEB@.RE;GET(ENTREEB);
                J:=FOUILLEBINAIRE(X);
                FICHTEMP@:=J;
                PUT(FICHTEMP);
                END;
        END;
ECRITCLASSES;
END;                (* CLASSE *)
 
PROCEDURE EUCLIDE;
VAR I,J,K,L,X:INTEGER;
    VERT:REAL;
BEGIN
REWRITE(FICHTEMP);
FOR I:=1 TO HAUTEUR DO
  FOR J:=1 TO LARGEUR DO
    FOR K:=1 TO HAUTEUR DO
        BEGIN
        VERT:=SQR(K-I);
        FOR L:=1 TO LARGEUR DO
                BEGIN
                X:=ROUND(SQRT(VERT+SQR(L-J)));
                FICHTEMP@:=X;
                PUT(FICHTEMP);
                END;
        END;
END;                (* ECHEC EUCLIDIEN *)
 
PROCEDURE LISMAT;
VAR I,J,X:INTEGER;
 
        PROCEDURE ENTRENEXT(I,J:INTEGER);
            BEGIN
        READ(CLASSEF,X);
        IF (X <=0)AND(I<>J) THEN AJOUTE(J,DEJAFAIT[I]);
        FICHTEMP@:=X;
        PUT(FICHTEMP);
        END;
BEGIN
NULL(DEJAFAIT);
IF LANG=1 THEN BEGIN
WRITELN(' OPTIONS d''entree pour la matrice de distances: ');
WRITELN('          1- Toute la matrice carree');
WRITELN('          2- La matrice triangulaire superieure seulement');
END ELSE BEGIN
WRITELN(' Input OPTIONS for distance matrix: ');;
WRITELN('          1- The entire square matrix');
WRITELN('          2- The upper triangular matrix only');
END;
WRITELN  ('         ');
VALIDE(I,1,2);
SYMMETRIC:=I=2;
RESET(CLASSEF);
REWRITE(FICHTEMP);
FOR I:=1 TO NOBJ DO
        BEGIN
        IF NOT(SYMMETRIC) THEN
                BEGIN
                FOR J:=1 TO I-1 DO ENTRENEXT(I,J);
                ENTRENEXT(I,I);
                END;
        FOR J:=I+1 TO NOBJ DO
                BEGIN
                ENTRENEXT(I,J);
                IF SYMMETRIC THEN IF X <=0 THEN AJOUTE(I,DEJAFAIT[J]);
                END;
        END;
END;                (* LISMAT *)
 
PROCEDURE LECTURE;
VAR I,J,K:INTEGER;
    X:REAL;
BEGIN
IF MOUVEMENT <> 15 THEN NULL(DIST1);
IF NOT(MOUVEMENT IN[12,13]) THEN
        FOR I:=1 TO LARGEUR*HAUTEUR DO READ(ENTREEC,MESURE[I])
ELSE
        BEGIN
        FOR I:=1 TO NOBJ DO READ(ENTREEC,COORDH[I],COORDV[I],MESURE[I]);
        END;
MOYENNE:=0;
FOR I:=1 TO NOBJ DO MOYENNE:=MOYENNE+MESURE[I];
MOYENNE:=MOYENNE/NOBJ;
VARIANCE:=0;
B2:=0;
FOR I:=1 TO NOBJ DO
        BEGIN
        X:=SQR(MESURE[I]-MOYENNE);
        VARIANCE:=VARIANCE+X;
        B2:=B2+SQR(X);
        END;
B2:=NOBJ*B2/SQR(VARIANCE);
END;                (* FIN LECTURE *)
 
 
PROCEDURE LECTCLASSE;
VAR I,J,K,L:INTEGER;
    X:REAL;
BEGIN
IF MOUVEMENT <> 15 THEN NULL(DIST1);
IF NOT(MOUVEMENT IN[12,13]) THEN
        FOR I:=1 TO NOBJ DO READ(ENTREEC,NOCLASSE[I])
ELSE FOR I:=1 TO NOBJ DO READ(ENTREEC,COORDH[I],COORDV[I],NOCLASSE[I]);
MAX:=0;
FOR I:=1 TO NOBJ DO
        BEGIN
        J:=NOCLASSE[I];
        IF J <= 0 THEN ERREUR(7,I)
        ELSE IF J > MAXCLASSES THEN ERREUR(8,I);
        IF J > MAX THEN MAX:=J;
        END;
FOR I:=1 TO MAX DO
        BEGIN
        K:=0;
        FOR J:=1 TO NOBJ DO IF NOCLASSE[J]=I THEN K:=K+1;
        MAT[I]:=K;
        END;
NRNS:=0;NR2NS2:=0;
FOR I:=1 TO MAX DO
        FOR J:=I-1 DOWNTO 1 DO
                BEGIN
                X:=MAT[I]*MAT[J];
                NRNS:=NRNS+X;
                NR2NS2:=NR2NS2+X*X;
                END;
NRNSNT:=0;
IF MAX >=3 THEN
    FOR I:=1 TO MAX-2 DO
        FOR J:=I+1 TO MAX-1 DO
            FOR K:=J+1 TO MAX DO NRNSNT:=NRNSNT+MAT[I]*MAT[J]*MAT[K];
NRNSNTNU:=0;
IF MAX >=4 THEN
    FOR I:=1 TO MAX-3 DO
        FOR J:=I+1 TO MAX-2 DO
            FOR K:=J+1 TO MAX-1 DO
                FOR L:=K+1 TO MAX DO
                    NRNSNTNU:=NRNSNTNU+MAT[I]*MAT[J]*MAT[K]*MAT[L];
NOBJREAL:=NOBJ;
N2:=NOBJREAL*(NOBJREAL-1.0);
N3:=N2*(NOBJREAL-2.0);
N4:=N3*(NOBJREAL-3.0);
END;                (* FIN LECTURE *)
 
PROCEDURE ECRITBITS;
VAR I,J:INTEGER;
BEGIN
(* To enable this debugging option, remove this line as well as the
   end-of-comment sign before the end of this procedure.
IF LANG=1 THEN
WRITELN (' Desirez-vous voir la matrice de bits',
         ' (fichier MATBIT) pour DEBUG ? ')
ELSE
WRITELN (' Do you want to see the matrix of bits',
         ' (file MATBIT) for debugging ?');
IF VALIDEBOOL THEN
        BEGIN
        REWRITE(MATBIT);
        FOR I:=1 TO NOBJ DO
                BEGIN
                WRITE(MATBIT,' ');
                FOR J:=1 TO NOBJ DO IF MEMBRE(J,DISTN[I]) THEN
                        WRITE(MATBIT,'1')ELSE WRITE(MATBIT,'0');
                WRITELN(MATBIT);
                END;
        END; *)
END;
 
PROCEDURE ECRITLIENS;
VAR I,J:INTEGER;
BEGIN
IF LANG=1 THEN
WRITELN(' Desirez-vous ecrire la liste des premiers liens sur ',
   'le fichier "LIENS" ?')
ELSE
WRITELN(' Do you wish to have the list of link edges written on ',
   'file "LINK" ?');
IF VALIDEBOOL THEN
        BEGIN
        LIENSECRITS:=TRUE;
        REWRITE(LIEN);
        NBLIENS:=0;
        FOR I:=1 TO NOBJ DO
                BEGIN
                FOR J:=1 TO NOBJ DO IF MEMBRE(J,DISTN[I]) THEN
                        BEGIN
                        NBLIENS:=NBLIENS+1;
                        WRITE(LIEN,'  ',I:4,' ',J:3);
                        IF NBLIENS MOD 8=0 THEN WRITELN(LIEN);
                        END;
                IF NBLIENS MOD 8 <> 0 THEN WRITELN(LIEN);
                END;
        END;
END;
 
PROCEDURE GABRIEL;
VAR I,J,K:INTEGER;
    DISTREF,X,Y:REAL;
 
         FUNCTION DISTANCE(I:INTEGER;X,Y:REAL):REAL;
        BEGIN
        DISTANCE:=SQRT(SQR(COORDH[I]-X)+SQR(COORDV[I]-Y));
        END;
 
BEGIN                (* GABRIEL *)
FOR I:=1 TO NOBJ DO
        BEGIN
        FOR J:=I+1 TO NOBJ DO
                BEGIN
                DISTREF:=DISTANCE(I,COORDH[J],COORDV[J])/2;
                IF DISTREF < EPSILON THEN ERREUR(1,J);
                K:=0;
                X:=(COORDH[I]+COORDH[J])/2;
                Y:=(COORDV[I]+COORDV[J])/2;
                COORDH[NOBJ+1]:=X;
                COORDV[NOBJ+1]:=Y;
                REPEAT
                        K:=K+1;
                UNTIL (DISTANCE(K,X,Y) < DISTREF)AND((K <> I)AND(K<>J));
                IF K > NOBJ THEN ADD(I,J,DIST1);
                END;
        END;
END;                (* FIN GABRIEL *)
 
PROCEDURE TESSELLATION(VAR CERCLE:CIRCLE;VAR DISTN:MATBITS);
   CONST EPS=10.0;
   VAR        I,J,K,L,NBTRIANGLES,DEBUT,NOBJCR,PTRYMIN,
              PTRYMAX,PTRXMIN,PTRXMAX:INTEGER;
        A,B,X1,Y1,X2,Y2,X3,Y3,CENTREX,CENTREY,RAYON,X,Y,X1RES,X2RES,
        SINTETA,COSTETA,SIN1TETA,COS1TETA:REAL;
        VEC:VECBITS;
        TROUVE,PLUSPETIT,INFINI:BOOLEAN;
 
        PROCEDURE CENTRE(VAR X1,Y1,X2,Y2,X3,Y3,CENTREX,CENTREY,
          RAYON2:REAL);
        VAR M1,M2,XPRIME,YPRIME,XX:REAL;
        BEGIN
        IF X2=X1 THEN X2:=X2+EPSILON;
        M1:=(Y2-Y1)/(X2-X1);
        IF X3=X1 THEN X3:=X3+EPSILON;
        M2:=(Y3-Y1)/(X3-X1);
        XPRIME:=(X3+X1)/2;
        YPRIME:=(Y3+Y1)/2;
        IF M2=0 THEN M2:=EPSILON;
        IF M1=0 THEN M1:=EPSILON;
        XX:=1/M2-1/M1;
        IF XX=0 THEN XX:=EPSILON;
        CENTREX:=(XPRIME/M2-(X2+X1)/(2*M1)+YPRIME-(Y2+Y1)/2)/XX;
        CENTREY:=(XPRIME-CENTREX)/M2+YPRIME;
        RAYON:=SQR(X1-CENTREX)+SQR(Y1-CENTREY);
        END;                (* CENTRE *)
 
        FUNCTION INTERSECTE(X,Y:REAL):BOOLEAN;
        BEGIN
        INTERSECTE:=SQR(X-CENTREX)+SQR(Y-CENTREY) < RAYON;
        END;                (* INTERSECTE *)
 
        PROCEDURE SIGNE(VAR A,B:REAL);
        BEGIN
        INFINI:=FALSE;
        IF ABS(X2-X1) >EPSILON THEN
                BEGIN
                A:=(Y2-Y1)/(X2-X1);
                B:=Y1-A*X1;
                PLUSPETIT:=Y3 > A*X3+B;
                END
        ELSE
                BEGIN
                INFINI:=TRUE;
                PLUSPETIT:=X2 < X3;
                END;
        END;
 
 
        PROCEDURE TOURNEANGLE(X,Y:REAL;VAR RESULTAT,SINA,COSA:REAL);
        BEGIN
        RESULTAT:=X*COSA+Y*SINA;
        END;
 
        PROCEDURE TROUVEANGLE(X,Y:REAL;VAR SINANGLE,COSANGLE:REAL);
        VAR XX:REAL;
        BEGIN
        IF ABS(X) <= EPSILON THEN BEGIN SINANGLE:=0;COSANGLE:=1;END
        ELSE
                BEGIN
                IF ABS(Y) < EPSILON THEN BEGIN SINANGLE:=1;
                COSANGLE:=0;END
                ELSE
                        BEGIN
                        XX:=ARCTAN(X/(-Y));
                        SINANGLE:=SIN(XX);
                        COSANGLE:=COS(XX);
                        (* SURTOUT NE PAS REMPLACER PAR SQRT() *)
                        END;
                END;
        END;
 
        FUNCTION CHERCHEINTER(X1,Y1,X2,Y2:REAL):BOOLEAN;
        VAR I:INTEGER;
            TROUVE:BOOLEAN;
 
                FUNCTION COMPRIS:BOOLEAN;
                BEGIN
                COMPRIS:=((X1RES > EPSILON) AND (X2RES < (-EPSILON)))
                OR
                ((X2RES > EPSILON) AND (X1RES < (-EPSILON)));
                END;
 
        BEGIN
        IF ELIMINATION THEN
          BEGIN
          I:=0;
          TROUVEANGLE(X2-X1,Y2-Y1,SINTETA,COSTETA);
          TROUVE:=FALSE;
          WHILE (I< NBTRIANGLES)AND(NOT(TROUVE)) DO
            BEGIN
            I:=I+1;
            WITH CERCLE[I] DO
              BEGIN
              TOURNEANGLE(COORDH[OBJ1]-X1,COORDV[OBJ1]-Y1,
                X1RES,SINTETA,COSTETA);
              TOURNEANGLE(COORDH[OBJ2]-X1,COORDV[OBJ2]-Y1,
                X2RES,SINTETA,COSTETA);
              IF COMPRIS THEN
                BEGIN
                TROUVEANGLE(COORDH[OBJ2]-COORDH[OBJ1],
                  COORDV[OBJ2]-COORDV[OBJ1],SIN1TETA,COS1TETA);
 
                TOURNEANGLE(X1-COORDH[OBJ1],Y1-COORDV[OBJ1],
                  X1RES,SIN1TETA,COS1TETA);
                TOURNEANGLE(X2-COORDH[OBJ1],
                  Y2-COORDV[OBJ1],X2RES,SIN1TETA,COS1TETA);
                TROUVE:=COMPRIS;
                END;
              END;
            END;
                CHERCHEINTER:=NOT(TROUVE);
                END
        ELSE CHERCHEINTER:=TRUE;
        END;                (* CHERCHEINTER *)
 
        FUNCTION MEMECOTE:BOOLEAN;
        VAR FX:REAL;
        BEGIN
        MEMECOTE:=FALSE;
        IF INFINI THEN
                BEGIN
                IF PLUSPETIT THEN MEMECOTE:=X < X2-EPSILON
                ELSE MEMECOTE:=X > X2+EPSILON;
                END
        ELSE
                BEGIN
                FX:=A*X+B;
                IF PLUSPETIT THEN MEMECOTE:=Y < FX-EPSILON
                ELSE MEMECOTE:=Y > FX+EPSILON;
                END;
        END;
 
        FUNCTION DIST2(X1,Y1,X2,Y2:REAL):REAL;
        BEGIN
        DIST2:=SQR(X1-X2)+SQR(Y1-Y2);
        END;                (* DIST2 *)
 
        PROCEDURE ENTRE(J,K,L:INTEGER);
        BEGIN
        IF NOT(MEMBRE(J,DISTN[K])) THEN
          BEGIN
          IF CHERCHEINTER(COORDH[J],COORDV[J],COORDH[K],COORDV[K]) THEN
                        BEGIN
                        NBTRIANGLES:=NBTRIANGLES+1;
                        IF NBTRIANGLES > MAXTRIANGLES THEN ERREUR(9,0);
                        WITH CERCLE[NBTRIANGLES] DO
                                BEGIN
                                OBJ1:=J;
                                OBJ2:=K;
                                OBJ3:=L;
                                END;
                        ADD(J,K,DISTN);
                        TROUVE:=TRUE;
                        END;
                END;
        END;                (* ENTRE *)
 
BEGIN                (* TESSELLATION *)
NBTRIANGLES:=0;
IF NBCR=-1 THEN
        BEGIN
        NBCR:=4;
        PTRYMIN:=1;PTRYMAX:=1;PTRXMIN:=1;PTRXMAX:=1;
        FOR I:=2 TO NOBJ DO
                BEGIN
                X:=COORDH[I];
                Y:=COORDV[I];
                IF X < COORDH[PTRXMIN] THEN PTRXMIN:=I
                ELSE IF X > COORDH[PTRXMAX] THEN PTRXMAX:=I;
                IF Y < COORDV[PTRYMIN] THEN PTRYMIN:=I
                ELSE IF Y > COORDV[PTRYMAX] THEN PTRYMAX:=I;
                END;
COORDH[NOBJ+1]:=COORDH[PTRXMIN]-EPS;COORDV[NOBJ+1]:=COORDV[PTRYMIN]-EPS;
COORDH[NOBJ+2]:=COORDH[PTRXMAX]+EPS;COORDV[NOBJ+2]:=COORDV[NOBJ+1];
COORDH[NOBJ+3]:=COORDH[NOBJ+1];     COORDV[NOBJ+3]:=COORDV[PTRYMAX]+EPS;
COORDH[NOBJ+4]:=COORDH[NOBJ+2];     COORDV[NOBJ+4]:=COORDV[NOBJ+3];
        END
ELSE FOR I:=NOBJ+1 TO NOBJ+NBCR DO
        BEGIN
        READ(ENTREEC,X1,Y1,X2,Y2);
        COORDH[I]:=(X1+X2)/2;
        COORDV[I]:=(Y1+Y2)/2;
        END;
NOBJCR:=NOBJ+NBCR;
NBMOTS:=(NOBJCR+1) DIV NBBITS;
NULL(DISTN);
FOR I:=NOBJ TO NOBJCR DO NULLVEC(DISTN[I+1]);
COORDH[NOBJCR+1]:=0;COORDV[NOBJCR+1]:=0;
DEBUT:=1;
RAYON:=MAXREEL;
X1:=COORDH[1];Y1:=COORDV[1];
FOR J:=2 TO NOBJ DO
        BEGIN
        X:=DIST2(X1,Y1,COORDH[J],COORDV[J]);
        IF X < RAYON THEN
                BEGIN
                K:=J;
                RAYON:=X;
                END;
        END;
X2:=COORDH[K];Y2:=COORDV[K];
IF K=2 THEN J:=3 ELSE J:=2;
X3:=COORDH[J];Y3:=COORDV[J];
ENTRE(1,K,J);
CENTRE(X1,Y1,X2,Y2,X3,Y3,CENTREX,CENTREY,RAYON);
REPEAT
        TROUVE:=FALSE;
        FOR I:=DEBUT TO NBTRIANGLES DO
                BEGIN
                WITH CERCLE[I] DO
                        BEGIN
                        X1:=COORDH[OBJ1];Y1:=COORDV[OBJ1];
                        X2:=COORDH[OBJ2];Y2:=COORDV[OBJ2];
                        X3:=COORDH[OBJ3];Y3:=COORDV[OBJ3];
                        SIGNE(A,B);
                        RAYON:=MAXREEL;
                        L:=0;K:=0;
                        REPEAT
                          L:=L+1;
                          IF L=OBJ1 THEN L:=L+1;
                          IF L=OBJ2 THEN L:=L+1;
                          IF L=OBJ3 THEN L:=L+1;
                          IF L > NOBJCR+1 THEN L:=NOBJCR+1;
                          X:=COORDH[L];Y:=COORDV[L];
                          IF MEMECOTE THEN
                            BEGIN
                            IF DIST2(CENTREX,CENTREY,X,Y) < RAYON THEN
                              BEGIN
                              K:=L;
                              X3:=X;Y3:=Y;
                              CENTRE(X1,Y1,X2,Y2,X3,Y3,CENTREX,
                                CENTREY,RAYON);
                              END;
                            END;
                        UNTIL (L > NOBJCR);
                        DEBUT:=DEBUT+1;
                        IF K<>0 THEN
                                BEGIN
                                ENTRE(OBJ1,K,OBJ2);
                                ENTRE(OBJ2,K,OBJ1);
                                END;
                        END;
                END;
UNTIL NOT(TROUVE);
FOR I:=NOBJ+1 TO NOBJCR+1 DO NULLVEC(DISTN[I]);
NULLVEC(VEC);
FOR I:=NOBJ+1 TO NOBJCR+1 DO AJOUTE(I,VEC);
FOR I:=1 TO NOBJ DO DIFFERENCE(DISTN[I],VEC);
END;                (* TESSELLATION *)
 
PROCEDURE OWN(NIVEAU:INTEGER);
VAR I,J,CLASSE:INTEGER;
BEGIN
RESET(FICHTEMP);
IF NIVEAU MOD 2=0 THEN DISTN:=DIST1
ELSE
   BEGIN
   NULL(DIST1);NULL(DISTN);
   IF (MOUVEMENT =0)OR((SYMMETRIC)AND(MOUVEMENT=14))
     THEN FOR I:=1 TO NOBJ DO
        BEGIN
        FOR J:=I+1 TO NOBJ DO
                BEGIN
                CLASSE:=FICHTEMP@;GET(FICHTEMP);
                IF CLASSE=NIVEAU THEN ADD(I,J,DISTN)
                ELSE IF CLASSE=NIVEAU+1 THEN ADD(I,J,DIST1);
                END;
        END
   ELSE FOR I:=1 TO NOBJ DO
            BEGIN
        FOR J:=1 TO NOBJ DO
                BEGIN
                CLASSE:=FICHTEMP@;GET(FICHTEMP);
                IF CLASSE=NIVEAU THEN AJOUTE(J,DISTN[I])
                ELSE IF CLASSE=NIVEAU+1 THEN AJOUTE(J,DIST1[I]);
                END;
        END;
   END;
END;                (* OWN *)
 
FUNCTION VARIMORAN:REAL;
BEGIN
VARIMORAN:=(NOBJREAL*((NOBJREAL*NOBJREAL-3*NOBJREAL+3)*S1-NOBJREAL*
  S2+3*W*W)-B2*((NOBJREAL*NOBJREAL-NOBJREAL)*S1-2*NOBJREAL*S2+6*W*W))
  /((NOBJREAL-1)*(NOBJREAL-2)*(NOBJREAL-3)*W*W)-1/SQR(NOBJREAL-1);
END;                (* VARIMORAN *)
 
FUNCTION VARIGEARY:REAL;
BEGIN
VARIGEARY:=((NOBJREAL-1)*S1*(NOBJREAL*NOBJREAL-3*NOBJREAL+3-
  (NOBJREAL-1)*B2)-0.25*(NOBJREAL-1)*S2*(NOBJREAL*NOBJREAL+3*NOBJREAL-6
  -(NOBJREAL*NOBJREAL-NOBJREAL+2)*B2)+W*W*(NOBJREAL*NOBJREAL-3-
  SQR(NOBJREAL-1)*B2))/(NOBJREAL*(NOBJREAL-2)*(NOBJREAL-3)*W*W);
END;
 
PROCEDURE RESULTATS(N:REAL);
VAR I,J,K,L,M,COMPTE,MIXEDTOTAL:INTEGER;
    COND,APPROPRIATE:BOOLEAN;
    SM,SC,VALEUR,EXPECTED,PROB,M2,TCG,TIM,TCOLOR,W2:REAL;
 
        FUNCTION MU2MEME(NR:REAL):REAL;
        VAR NR2,NR3,NR4:REAL;
        BEGIN
        NR2:=NR  *(NR-1.0);
        NR3:=NR2*(NR-2.0);
        NR4:=NR3*(NR-3.0);
        MU2MEME:= ( (S1*NR2/N2)+(S2-2*S1)*NR3/N3 +
                (W2+S1-S2)*NR4/N4 - W2*NR2*NR2/N2/N2)/4;
        END;
 
        FUNCTION MU2DIFF(NR,NS:REAL):REAL;
        VAR NRNS:REAL;
        BEGIN
        NRNS:=NR*NS;
        MU2DIFF:=(2*S1*NRNS/N2+(S2-2*S1)*NRNS*(NR+NS-2)/N3
                +4*(W2+S1-S2)*NRNS*(NR-1)*(NS-1)/N4
                -4*W2*SQR(NRNS/N2))/4;
        END;        (* MU2DIFF *)
 
        FUNCTION MU2TOTAL:REAL;
        BEGIN
        MU2TOTAL:=((S2/N2-4*(W2+S1-S2)*(N-1)/N4)*NRNS
                +((2*S1-5*S2)/N3+12*(W2+S1-S2)/N4+8*W2/N3/(N-1))*NRNSNT
                +((S1-S2)/N4+2*W2*(2*N-3)/N2/N4)*
                (NR2NS2*4-NRNSNTNU*8))/4;
        END;   (* MU2TOTAL *)
 
        FUNCTION TROUVEALPHA(X,CORRECTION:REAL;OPERATION:INTEGER):REAL;
        VAR Z,ZP,ZM,MEDIUM:REAL;
 
                FUNCTION F(ALPHA:REAL):REAL;
                VAR Y:REAL;
                BEGIN
                Z:=NORMDEV(ALPHA);
                IF X > 0 THEN ALPHA:=1-ALPHA;
                Y:=SQRT(10*ALPHA);
                CASE OPERATION OF
                        1:F:=Z-Y*CORRECTION;
                        2:F:=Z+Y*CORRECTION;
                        3:F:=Z-CORRECTION*(1-Y);
                        END;        (* CASE *)
                END;
 
        BEGIN                (* TROUVEALPHA *)
        ZP:=1;
        ZM:=0;;
        WHILE ABS(ZP-ZM) > PROBPREC DO
                BEGIN
                MEDIUM:=(ZP+ZM)/2;
                IF F(MEDIUM) > X THEN ZP:=MEDIUM ELSE ZM:=MEDIUM;
                END;
        TROUVEALPHA:=MEDIUM;
        END;                (* TROUVEALPHA *)
 
        PROCEDURE CHERCHEPROB;
        BEGIN
        IF COND THEN
                BEGIN
                M2:=SQRT(M2);
                IF VALEUR < 0 THEN PROB:=TROUVEALPHA(VALEUR,3/M2,3)
                ELSE PROB:=TROUVEALPHA(VALEUR,1/M2,3);
                END
        ELSE PROB:=GAUSS(VALEUR);
        END;
 
        PROCEDURE ECRITPROB;
        BEGIN
        WRITE(SORTIEC,VALEUR:11:3);
        IF VALEUR > 0 THEN WRITE(SORTIEC,1-PROB:14:3,' ':14)
        ELSE WRITE(SORTIEC,' ':14,PROB:14:3);
        END;
BEGIN                (* RESULTATS *)
IF W > 0 THEN
    BEGIN
    COND:=(N <= 50)AND(NOT(PGTRENTEPOURCENT)) ;
    KALPHA:=1;
    K:=ROUND(W);
    TOTAL:=TOTAL+K;
    IF NOMINAL THEN
        BEGIN
        WRITELN(SORTIEC,' DISTANCE',NIVEAU:3,K:52);
        W2:=W*W;
        MIXEDTOTAL:=0;
        FOR I:=1 TO MAX DO
            FOR J:=I TO MAX DO
                BEGIN
                APPROPRIATE:=TRUE;
                PROB:=MAT[I];
                IF I <> J THEN EXPECTED:=W*PROB*MAT[J]/N2
                ELSE
                  BEGIN
                  EXPECTED:=MAT[I]/NOBJ;
                  IF (EXPECTED < 0.2)OR(EXPECTED > 0.8)
                    THEN APPROPRIATE:=FALSE;
                  EXPECTED:=W*PROB*(PROB-1.0)/(2*N2);
                  END;
                COMPTE:=0;
                WRITE(SORTIEC,' ':12,'[',I:1,'][',J:1,']');
                FOR L:=1 TO NOBJ DO
                  FOR M:=L+1 TO NOBJ DO
                    IF (MEMBRE(L,DISTN[M]))OR(MEMBRE(M,DISTN[L])) THEN
                      BEGIN
                      IF ((NOCLASSE[L]=I)AND(NOCLASSE[M]=J))
                        OR((NOCLASSE[L]=J)AND(NOCLASSE[M]=I))THEN
                                        COMPTE:=COMPTE+1;
                                END;
                IF I=J THEN M2:=MU2MEME(PROB)
                ELSE
                  BEGIN
                  MIXEDTOTAL:=MIXEDTOTAL+COMPTE;
                  M2:=MU2DIFF(PROB,MAT[J]);
                        END;
                IF APPROPRIATE THEN
                    BEGIN
                    IF M2 >EPSILON THEN
                        BEGIN
                        VALEUR:=(COMPTE-EXPECTED)/SQRT(M2);
                        IF I=J THEN PROB:=1-CHIPROB(COMPTE*2*
                          EXPECTED/M2,TRUNC(2*EXPECTED*EXPECTED/M2))
                        ELSE CHERCHEPROB;
                        ECRITPROB;
                        END
                    ELSE IF LANG=1 THEN
                      WRITE(' Calcul impossible, car variance <= 0')
                      ELSE
                      WRITE(' Cannot compute, since variance <= 0');
                    END
    ELSE
        IF LANG=1 THEN WRITE
        (SORTIEC,' CARD. CLASSE [',I:1,']/NOBJ < 0.2 ou > 0.8')
        ELSE WRITE
        (SORTIEC,' CARD. CLASS [',I:1,']/NOBJ < 0.2 or > 0.8');
                WRITELN(SORTIEC);
                END;
        WRITE(SORTIEC,'     [Total diff.]');
        M2:=MU2TOTAL;
        VALEUR:=(MIXEDTOTAL-W*NRNS/N2)/SQRT(M2);
        CHERCHEPROB;
        ECRITPROB;
        WRITELN(SORTIEC);
        END
    ELSE
        BEGIN
        IF COND THEN COND:=(N-SQRT(N)<W/4)
                        AND(W/4<=(2.0*N-3.0*SQRT(N)+1.0));
        IMORAN:=N*IMORAN/(VARIANCE*W);
        CGEARY:=(N-1)*CGEARY/(2*W*VARIANCE);
        SC:=SQRT(VARIGEARY);
        SM:=SQRT(VARIMORAN);
        IF (SC=0)OR(SM=0) THEN ERREUR(10,0);
        TCG:=(1-CGEARY)/SC;
        TIM:=IMORAN/SM;
        IF COND THEN
                BEGIN
                PROBI:=TROUVEALPHA(TIM,1/(N-1)/SM,1);
                IF CGEARY > 1 THEN PROBC:=TROUVEALPHA(TCG,1/(N-1)/SC,2)
                ELSE PROBC:=GAUSS(TCG);
                END
        ELSE
                BEGIN
                PROBI:=GAUSS(TIM+1/(N-1)/SM);
                IF CGEARY > 1 THEN PROBC:=GAUSS(TCG-1/(N-1)/SC)
                ELSE PROBC:=GAUSS(TCG);
                END;
        WRITE(SORTIEC,NIVEAU:4,IMORAN:9:4);
        IF IMORAN > 0 THEN WRITE(SORTIEC,1-PROBI:9:3,' ':9)
        ELSE WRITE(SORTIEC,' ':9,PROBI:9:3);
        WRITE(SORTIEC,CGEARY:9:4);
        IF CGEARY < 1 THEN WRITE(SORTIEC,1-PROBC:9:3,' ':9)
        ELSE WRITE(SORTIEC,' ':9,PROBC:9:3);
        WRITELN(SORTIEC,K:6);
        END;
    END;
END;        (* RESULTATS *)
 
PROCEDURE CALCUL;
VAR I,J,K:INTEGER;
    MES,MESLESSMU,WV,WH:REAL;
    WIJ:BOOLEAN;
BEGIN
S1:=0;S2:=0;W:=0;IMORAN:=0;CGEARY:=0;
PGTRENTEPOURCENT:=FALSE;
FOR I:=1 TO NOBJ DO
        BEGIN
        IF NOT(NOMINAL) THEN
                BEGIN
                MES:=MESURE[I];
                MESLESSMU:=MES-MOYENNE;
                END;
        WH:=0;WV:=0;
        FOR J:=1 TO NOBJ DO
                BEGIN
                WIJ:=FALSE;
                IF MEMBRE(I,DISTN[J])THEN
                        BEGIN
                        WIJ:=TRUE;
                        WV:=WV+1;
                        END;
                IF MEMBRE(J,DISTN[I]) THEN
                        BEGIN
                        IF WIJ THEN S1:=S1+4 ELSE S1:=S1+1;
                        W:=W+1;
                        WH:=WH+1;
                        IF NOT(NOMINAL) THEN
                          BEGIN
                          IMORAN:=IMORAN+(MESURE[J]-MOYENNE)*MESLESSMU;
                          CGEARY:=CGEARY+SQR(MESURE[J]-MES);
                          END;
                        END;
                END;
        WH:=WH+WV;
        S2:=S2+SQR(WH);
        PGTRENTEPOURCENT:=WH > TRENTEPOURCENT;
        END;
S1:=S1/2;
RESULTATS(NOBJREAL);
END;                (* CALCUL *)
 
PROCEDURE AUTOCORRELATION;
VAR I,J,NBPOSSIBLES,ANCIENCARDI:INTEGER;
    ACTION:BOOLEAN;
 
        PROCEDURE NEXT(VAR DIST1,DISTN,DEJAFAIT:MATBITS);
        VAR I,J,CARDI:INTEGER;
        BEGIN
        IF MOUVEMENT IN [1..10,12,13,15] THEN
                BEGIN
                FOR I:=1 TO NOBJ DO
                        BEGIN
                        TEMP:=DISTN[I];
                        FOR J:=1 TO NOBJ DO IF MEMBRE(J,TEMP) THEN
                                UNION(DISTN[I],DIST1[J]);
                        END;
                FOR I:=1 TO NOBJ DO DIFFERENCE(DISTN[I],DEJAFAIT[I]);
                END
        ELSE IF MOUVEMENT IN[0,11,14] THEN OWN(NIVEAU+1);
        UNIONMAT(DEJAFAIT,DISTN,CARDI);
        IF MOUVEMENT=9 THEN CARDI:=CARDI+CARDI;
        IF (CARDI=NBPOSSIBLES)
           OR((MOUVEMENT IN[1..10,12,13,15])AND
           (CARDI=ANCIENCARDI)) THEN ACTION:=FALSE;
        ANCIENCARDI:=CARDI;
        END;
 
BEGIN        (* AUTOCORRELATION *)
ANCIENCARDI:=0;
NIVEAU:=1;
IF MOUVEMENT IN[1..7] THEN UNIONMAT(DEJAFAIT,DIST1,I)
ELSE IF MOUVEMENT <> 14 THEN DEJAFAIT:=DIST1;
IF MOUVEMENT IN[0,11,14] THEN
        BEGIN
        OWN(1);
        IF MOUVEMENT <> 14 THEN DEJAFAIT:=DISTN
        ELSE UNIONMAT(DEJAFAIT,DISTN,I);
        END
ELSE DISTN:=DIST1;
IF (MOUVEMENT <>14)OR(SYMMETRIC) THEN FOR I:=1 TO NOBJ
   DO AJOUTE(I,DEJAFAIT[I]);
LIENSECRITS:=FALSE;
IF MOUVEMENT <> 15 THEN ECRITLIENS;
ECRITBITS;
IF NOMINAL THEN
      BEGIN
      WRITELN(SORTIEC,' H0:',' ':32,'S.N.D.=0,    S.N.D.=0');
      WRITELN(SORTIEC,' H1:',' ':32,'S.N.D.>0,    S.N.D.<0');
      IF LANG=1 THEN
      WRITELN(SORTIEC,' ':11,'CLASSES     S.N.D.         P(H0) ,',
      '       P(H0), PAIRES')
      ELSE
      WRITELN(SORTIEC,' ':11,'CLASSES     S.N.D.         P(H0) ,',
      '       P(H0),  PAIRS');
      END
ELSE
      BEGIN
      WRITELN(SORTIEC,' H0:',' ':13,'I = 0    I = 0',' ':13,'C = 1',
      '    C = 1');
      WRITELN(SORTIEC,' H1:',' ':13,'I > 0    I < 0',' ':13,'C < 1',
      '    C > 1');
      IF LANG=1 THEN
      WRITELN(SORTIEC,' DIST.,I(MORAN), P(H0),   P(H0), C(GEARY),',
      '  P(H0),   P(H0), PAIRES')
      ELSE
      WRITELN(SORTIEC,' DIST.,I(MORAN), P(H0),   P(H0), C(GEARY),',
      '  P(H0),   P(H0),  PAIRS');
      END;
NBPOSSIBLES:=NOBJ*NOBJ;
ACTION:=TRUE;
NOBJREAL:=NOBJ;
TRENTEPOURCENT:=NOBJREAL*(NOBJREAL-1)*3/10;
CALCUL;
REPEAT
        NEXT(DIST1,DISTN,DEJAFAIT);
        NIVEAU:=NIVEAU+1;
        CALCUL;
UNTIL NOT(ACTION);
END;        (* AUTOCORRELATION *)
 
PROCEDURE ECHECS(VAR DISTANCE,DEJAFAIT:MATBITS);
VAR I,J,K:INTEGER;
 
        PROCEDURE INITBIT;
        VAR I,J:INTEGER;
        BEGIN
        FOR I:=0 TO NBMOTS DO DEJAFAIT[1,I]:=[];
        FOR I:=1 TO NOBJ DO
                BEGIN
                J:=I DIV NBBITS;
                DEJAFAIT[1,J]:=DEJAFAIT[1,J]+[I MOD NBBITS];
                END;
        FOR I:=2 TO NOBJ DO DEJAFAIT[I]:=DEJAFAIT[1];
        END;   (* INITBIT *)
 
        PROCEDURE VALIDE(V,H,N:INTEGER);
        BEGIN
        IF (H >=1) AND (H<=LARGEUR) THEN
           IF (V>=1) AND (V<=HAUTEUR) THEN ADD((V-1)*LARGEUR+H,N,
           DISTANCE);
        END;
 
        PROCEDURE TOURH(DIRECT:BOOLEAN);
        VAR I,J,LIMINF:INTEGER;
        BEGIN
        LIMINF:=0;
        FOR I:=1 TO HAUTEUR DO
                BEGIN
                FOR J:=1 TO LARGEUR DO
                        BEGIN
                        LIMINF:=LIMINF+1;
                        VALIDE(I,J+1,LIMINF);
                        VALIDE(I,J-1,LIMINF);
                        END;
                END;
        IF DIRECT THEN
          BEGIN
          LIMINF:=1;
          FOR I:=1 TO NOBJ DO
            BEGIN
            FOR J:=LIMINF+LARGEUR-1 DOWNTO LIMINF DO SUB(I,J,DEJAFAIT);
                        IF I MOD LARGEUR=0 THEN LIMINF:=LIMINF+LARGEUR;
                        END;
                END;
        END;
 
        PROCEDURE TOURV(DIRECT:BOOLEAN);
        VAR I,J,P:INTEGER;
            TEMP:VECBITS;
        BEGIN
        P:=0;
        FOR I:=1 TO HAUTEUR DO
                BEGIN
                FOR J:=1 TO LARGEUR DO
                        BEGIN
                        P:=P+1;
                        VALIDE(I+1,J,P);
                        VALIDE(I-1,J,P);
                        END;
                END;
        IF DIRECT THEN
                BEGIN
                FOR I:=1 TO LARGEUR DO
                        BEGIN
                        P:=I;
                        FOR J:=1 TO HAUTEUR DO
                                BEGIN
                                ENLEVE(P,DEJAFAIT[I]);
                                P:=P+LARGEUR;
                                END;
                        P:=I;
                        FOR J:=1 TO HAUTEUR DO
                                BEGIN
                                DEJAFAIT[P]:=DEJAFAIT[I];
                                P:=P+LARGEUR;
                                END;
                        END;
                END;
        END;
 
        PROCEDURE FOU(POSITIF,DIRECT:BOOLEAN);
        VAR I,J,N,N1,H,V,SENS:INTEGER;
        BEGIN
        IF POSITIF THEN SENS:=-1 ELSE SENS:=1;
        N:=0;
        FOR I:=1 TO HAUTEUR DO
                BEGIN
                FOR J:=1 TO LARGEUR DO
                        BEGIN
                        N:=N+1;
                        VALIDE(I+1,J+SENS,N);
                        VALIDE(I-1,J-SENS,N);
                        END;
                END;
        IF DIRECT THEN
                BEGIN
                N:=1;
                FOR I:=1 TO HAUTEUR DO
                   BEGIN
                   FOR J:=1 TO LARGEUR DO
                     BEGIN
                     V:=I-1;
                     H:=J-SENS;
                     N1:=N-LARGEUR-SENS;
                     WHILE (V > 0)AND((H >=1)AND(H<=LARGEUR)) DO
                                        BEGIN
                                        SUB(N1,N,DEJAFAIT);
                                        V:=V-1;
                                        H:=H-SENS;
                                        N1:=N1-LARGEUR-SENS;
                                        END;
                      V:=I+1;
                      H:=J+SENS;
                      N1:=N+LARGEUR+SENS;
                      WHILE (V <= HAUTEUR)AND((H>=1)AND(H<=LARGEUR)) DO
                                        BEGIN
                                        SUB(N1,N,DEJAFAIT);
                                        V:=V+1;
                                        H:=H+SENS;
                                        N1:=N1+LARGEUR+SENS;
                                        END;
                                N:=N+1;
                                END;
                        END;
                END;
        END;
 
BEGIN        (* MOUVEMENT DES ECHECS *)
INITBIT;
CASE MOUVEMENT OF
        1:TOURH(TRUE);
        2:TOURV(TRUE);
        3:FOU(TRUE,TRUE);
        4:FOU(FALSE,TRUE);
        5:BEGIN
          TOURV(TRUE);
          TOURH(TRUE);
          END;
        6:BEGIN
          FOU(TRUE,TRUE);
          FOU(FALSE,TRUE);
          END;
        7:BEGIN                (* REINE *)
          TOURV(TRUE);
          TOURH(TRUE);
          FOU(TRUE,TRUE);
          FOU(FALSE,TRUE);
          END;
        8:BEGIN
          TOURV(FALSE);
          TOURH(FALSE);
          END;   (* MODIFIER *)
        9:BEGIN
          FOU(TRUE,FALSE);
          FOU(FALSE,FALSE);
          END;
        10:BEGIN        (* REINE *)
          TOURV(FALSE);
          TOURH(FALSE);
          FOU(TRUE,FALSE);
          FOU(FALSE,FALSE);
          END;
        END;        (* CASE *)
END;                (* ECHECS *)
 
 
BEGIN                (* PROG PRINC *)
(* TERMOUT(OUTPUT); *) (* Not needed; may cause strange EXEC behavior *)
INIT;
PARAMETRES;
IF MOUVEMENT=0 THEN
        BEGIN
        NEW(PTRREAL);
        NEW(FREQUENCES);
        NEW(DONNEES);
        ENTETE(ASAUTER);
        HAUTEUR:=1;
        LARGEUR:=NOBJ;
        NBMOTS:=NOBJ DIV NBBITS;
        CLASSE(DONNEES@,PTRREAL@,FREQUENCES@);
        DISPOSE(PTRREAL);
        DISPOSE(FREQUENCES);
        DISPOSE(DONNEES);
        END;
IF LANG=1 THEN
WRITELN(SORTIEC,' Option du mouvement: ',MOUVEMENT:4)
ELSE
WRITELN(SORTIEC,' Movement option: ',MOUVEMENT:4);
WRITELN(SORTIEC);
IF LANG=1 THEN BEGIN
WRITELN(SORTIEC,' NOTE: ',
  'Les probabilites les plus significatives sont pres de zero');
WRITELN(SORTIEC,' Les probabilites sont imprimees a la precision de',
  PROBPREC:10:5);
END ELSE BEGIN
WRITELN(SORTIEC,' NOTE: ',
  'The most significant probabilities are close to zero');
WRITELN(SORTIEC,' The probabilities are printed with precision',
  PROBPREC:10:5);
END;
WRITELN(SORTIEC);
IF NOMINAL THEN LECTCLASSE
ELSE LECTURE;
CASE MOUVEMENT OF
        12,13:        BEGIN
                IF MOUVEMENT=12 THEN GABRIEL
                ELSE
                        BEGIN
                        NEW(CERCLE);
                        TESSELLATION(CERCLE@,DIST1);
                        DISPOSE(CERCLE);
                        END;
                END;
        1,2,3,4,5,6,7,8,9,10:ECHECS(DIST1,DEJAFAIT);
        11:EUCLIDE;
        0,15:BEGIN END;
        14:LISMAT;
        END;        (* CASE *)
AUTOCORRELATION;
WRITELN(SORTIEC,' TOTAL',' ':50,TOTAL:8);
IF LIENSECRITS THEN BEGIN
   IF LANG=1 THEN WRITELN
(' ***',NBLIENS:5,' liens ont ete ecrits sur le fichier de LIENS ***')
   ELSE WRITELN
(' ***',NBLIENS:5,' links have been written onto the LINK file ***');
END;
777: IF LANG=1 THEN WRITELN(' Fin du programme.  ')
               ELSE WRITELN(' End of the program.');
END.
