PROGRAM KMEANS(Input,Output,Entree,Fichclas,Sortie,Lien,Coordxy);
const maxvar   =12;    (* maximum number of variables                 *)
      maxobj   =800;   (* maximum number of objects                   *)
      maxgrp   =30;    (* maximum number of groups                    *)
      epsilon  =10e-7;
      nbbits   =50;    (* maximum length of a Pascal set              *)
      nbbitsl1 =49;    (* nbbits -1                                   *)
      maxmots  =16;    (* maxobj / nbbits                             *)
      facteur  = 1;    (* printing map factor, Y/X                    *)
      lang     =1;     (* parametre de langue:  1=Francais
                          language parameter :  2=English             *)

(* ================================================================== *)
(* This program computes K-means clustering, or minimum-variance      *)
(* hierarchical clustering (Ward 1963), by one of the following       *)
(* methods:                                                           *)
(*                                                                    *)
(* 1) Unconstrained clustering, following Ward (1963) and             *)
(*    Spath (1975, 1980).                                             *)
(*                                                                    *)
(* 1a) The Stony Brook method: perform N iterations, starting         *)
(*     every time with a different random configuration. The solution *)
(*     that minimizes a statistic D is kept, where D is the sum, for  *)
(*     all groups J, of the sum of the squares of the distances       *)
(*     between the members of the J-th cluster and their centroid     *)
(*     (Spath 1980, page 73).                                         *)
(* 1b) Enter your own initial configuration, issued for instance from *)
(*     another clustering or ordination technique. This configuration *)
(*     is entered on file FICHCLAS.                                   *)
(*     This seems to be the best method to prevent from getting stuck *)
(*     into a local minimum of the D function.                        *)
(* 1c) The Spath method (1980, page 67), also called the MODULUS      *)
(*     method, that consists of attributing object 1 to group 1,      *)
(*     object 2 to group 2, ... , object k to group k, object k+1 to  *)
(*     group 1, etc.This is taken as the initial configuration.       *)
(*                                                                    *)
(* 2) Clustering with contiguity constraint.                          *)
(*                                                                    *)
(*   - The constraint can be one-dimensional, as in Chronological     *)
(*     Clustering, in which case the program will assume that succes- *)
(*     sive objects in the data list are adjacent;                    *)
(*   - Or, the constraints (spatial or other) imply a space with two  *)
(*     or more dimensions. In this case, the constraints are entered  *)
(*     as a list of connections on file LIEN, just as in BIOGEO       *)
(*     or in PRECEL.                                                  *)
(*                                                                    *)
(*     When clustering with constraint, the choice for the starting   *)
(* configurations is between                                          *)
(*                                                                    *)
(* 2a) The Stony Brook method (as in 1a above), or                    *)
(* 2b) Your own initial configuration (as in 1b above).               *)
(*                                                                    *)
(* Results are presented, either for the initial and final configur-  *)
(* ation only, or for all intermediate steps. They include the group  *)
(* membership, the statistic E for each group (E is the within-group  *)
(* sum of squared distances from the points to the centroid), and the *)
(* D statistic for the whole solution (D is the sum of the E's.)      *)
(*                                                                    *)
(* Addition by Esther Beneish, July 1995: the program can now draw    *)
(* in the case of constrained clustering.                             *)
(* ================================================================== *)
(* Files:                                                             *)
(*    INPUT and OUTPUT are connected to the user's terminal.          *)
(*    ENTREE is the raw data matrix. It consists of a rectangular     *)
(*       array of objects (lines) by variables (columns).             *)
(*    FICHCLAS contains the starting configuration(s), if any. It is  *)
(*       written as follows.                                          *)
(*       For the sake of the example, suppose that you want to make   *)
(*       2 iterations, and that there are 4 groups and 13 objects.    *)
(*       The file might look like this:                               *)
(*         1 7 3 12 0                                                 *)
(*         8 2 0                                                      *)
(*         10 13 4 5 0                                                *)
(*         6 9 11 0       up to now, this is for the first iteration; *)
(*         2 4 10 0                                                   *)
(*         9 1 3 13 0                                                 *)
(*         12 5 6 0                                                   *)
(*         7 8 11 0       which completes the group assignment.       *)
(*       Notice that each group ends with a zero. If an object has    *)
(*       not been assigned to a group, the program will ask for the   *)
(*       group assignment of that object. When an object has been     *)
(*       assigned to several groups, the last assignment overrides    *)
(*       the previous ones.                                           *)
(*                                                                    *)
(*    SORTIE is the file containing the results.                      *)
(*    LIEN is the file of link edges. It may have been produced by    *)
(*       another program, like AUTOCOR. Or, the user can write his    *)
(*       own file of link edges directly; it is then a simple list    *)
(*       of pairs of points.                                          *)
(*    COORDXY is a file of coordinates. It is used to draw maps.      *)
(*                                                                    *)
(* This program uses Function GGUBFS for generating random numbers.   *)
(* The user can take either the GGUBFS Function included in this      *)
(* program, or attach the GGUBFS Function of IMSL. A call to the      *)
(* IMSL version of the function is provided as a comment.             *)
(* ================================================================== *)
type  statistyp=array[1..maxobj,1..maxvar] of real;
      ecarttyp =array[1..maxgrp,1..maxvar] of real;
      position =array[1..maxobj] of integer;
      contenu  =array[1..maxgrp] of integer;
      grbool   =array[1..maxgrp] of boolean;
      ensemble=set of 0..nbbitsl1;
      vecteurbit=array[0..maxmots]of ensemble;
      neighbours=array[1..maxobj] of vecteurbit;
      matbit= array[1..maxgrp] of vecteurbit;
      rec=record re1,re2:real;end;
      memtype=array[1..maxobj] of integer;
      coordtype=array[1..2,1..maxobj] of integer;
var   nobj,nbvar,nbgroupes,nbmots,typedegroupement,
      nbessais,noboucle,option,largeur:integer;
      dseed:real;
      fichierdegroupes,intermediaire,graphe,change:boolean;
      donnees:statistyp;
      moyenne:ecarttyp;
      posgroupes:position;
      cardinalite:contenu;
      voisins:neighbours;
      groupes,frontiere:matbit;
      entree,fichclas,sortie,lien,coordxy:text;
      trouve:grbool;

procedure erreur(i,j:integer);
begin
if lang=1 then case i of
      1:writeln(' Numero de groupe > nombre de groupes');
      2:writeln(' Numero d''objet  > nombre d''objets ');
      3:writeln(' Pas de variation dans la variable ',j:1);
      4:writeln(' Objet ',j:1,' deja alloue a un groupe');
      5:writeln(' Fichier de liens: objet(s) non connecte(s)');
      end
else case i of
      1:writeln(' Group number > number of groups');
      2:writeln(' Object number > number of objects');
      3:writeln(' No variation for variable ',j:1);
      4:writeln(' Object ',j:1,' already alloted to a group');
      5:writeln(' File of link edges: there are unconnected object(s)');
      end;
end;

procedure valide(limg,limd:integer;var x:integer);
var fait:boolean;
begin
repeat
      fait:=true;
      readln (x);
      if (x < limg) or (x > limd) then
            begin
            fait:=false;
            if lang=1 then writeln(' S.V.P. ',limg:1,'..',limd:1)
               else writeln(' PLEASE ',limg:1,'..',limd:1);
            end;
until fait;
end;

Function GGUBFS (var dseed:real):real;
const d2p31m=2147483647.0;
      d2p31 =2147483648.0;
var a:real;
begin
a:=16807*dseed;
dseed:=a-trunc(a/d2p31m)*d2p31m;
ggubfs:=dseed/d2p31;
end;                        (* end of the PASCAL version of GGUBFS, May 1987 *)

procedure geninit;
var i,j:integer;
    x:real;
begin
if lang=1 then
     begin
     writeln(' Tapez un chiffre (petit entier) pour indiquer le ');
     writeln(' point de depart du generateur de nombres aleatoires ');
     end
else writeln(' Type a small integer for the random number generator ');
readln(i);
for j:=1 to i do x:=ggubfs(dseed);
end;

function card(x:ensemble):integer;
var c,i:integer;
begin
c:=0;
if x <> [] then for i:=0 to nbbitsl1 do if i in x then c:=c+1;
card:=c;
end;            (* card *)

function vide(var m:vecteurbit):boolean;
var i:integer;
    b:boolean;
begin
b:=true;
for i:=0 to nbmots do if m[i]<>[] then b:=false;
vide:=b;
end;

function membre(ind:integer;var mm:vecteurbit):boolean;
var i,j:integer;
begin
i:=ind div nbbits;
j:=ind mod nbbits;
membre:=j in mm[i];
end;

procedure nullvec(var m:vecteurbit);
var i:integer;
begin
for i:=0 to nbmots do m[i]:=[];
end;

procedure union(var m1,m2:vecteurbit);
var i:integer;
begin
for i:=0 to nbmots do m1[i]:=m1[i]+m2[i];
end;

procedure intersection(var m0,m1,m2:vecteurbit);
var i:integer;
begin
for i:=0 to nbmots do m0[i]:=m1[i]*m2[i];
end;

function inclusegal(var m1,m2:vecteurbit):boolean;
var i:integer;
    b:boolean;
begin
b:=true;
for i:=0 to nbmots do if not(m1[i] <= m2[i]) then b:=false;
inclusegal:=b;
end;

function egal(var m1,m2:vecteurbit):boolean;
var i:integer;
    b:boolean;
begin
b:=true;
for i:=0 to nbmots do if m1[i] <> m2[i] then b:=false;
egal:=b;
end;

procedure ajoute(ind:integer;var m:vecteurbit);
var i:integer;
begin
i:=ind div nbbits;
m[i]:=m[i]+[ind mod nbbits];
end;

procedure enleve(ind:integer;var m:vecteurbit);
var i:integer;
begin
i:=ind div nbbits;
m[i]:=m[i]-[ind mod nbbits];
end;

procedure difference(var m1,m2:vecteurbit);
var i:integer;
begin
for i:=0 to nbmots do m1[i]:=m1[i]-m2[i];
end;

function cardvect(var mat:vecteurbit):integer;
var i,compte:integer;
begin
compte:=0;
for i:=0 to nbmots do compte:=compte+card(mat[i]);
cardvect:=compte;
end;

procedure creefrontiere(var frontiere,carte:vecteurbit);
var i:integer;
begin
nullvec(frontiere);
for i:=1 to nobj do if membre(i,carte) then union(frontiere,voisins[i]);
difference(frontiere,carte);
end;

procedure premier(var i:integer;var t:vecteurbit);
begin
i:=0;
repeat i:=i+1; until membre(i,t);
end;

function connexe:boolean;
var i,j,nbi:integer;
    fait,carte,frontiere:vecteurbit;
    connecte:boolean;

    procedure connexion(var mat:vecteurbit);
    var i,j:integer;
    begin
    if vide(mat) then connecte:=false
    else
      begin
      i:=0;
      premier(i,mat);
      nullvec(fait);
      ajoute(i,fait);
      carte:=fait;
      repeat
            creefrontiere(frontiere,carte);
            carte:=frontiere;
            difference(carte,fait);
            intersection(carte,carte,mat);
            union(fait,carte);
            if vide(carte) then connecte:=false;
      until (egal(mat,fait)) or (not(connecte));
      end;
    end;

begin
connecte:=true;
for i:=1 to nbgroupes do if connecte then
      begin
      nbi:=0;
      for j:=1 to nobj do if membre(j,groupes[i]) then nbi:=nbi+1;
      if nbi > 1 then connexion(groupes[i]);
      end;
connexe:=connecte;
end;

procedure resultats(prof:integer);
var i,j,k,count,cardiint:integer;
    e,sume,cardireel:real;
begin
writeln(sortie);
if lang=1 then write(sortie,'   Etape ')
          else write(sortie,'   Configuration:  ');
if prof= 0 then
      begin
      if lang=1 then writeln(sortie,'initiale')
                else writeln(sortie,'initial');
      end
else
      begin
      if prof < 0 then
           begin
           if lang=1 then writeln(sortie,'finale (',- prof:1,' etapes)')
              else writeln(sortie,'final (',- prof:1,' steps)');
           end
      else writeln(sortie,prof:1);
      end;
sume:=0;
for i:=1 to nbgroupes do
      begin
      if lang=1 then write(sortie,'     Groupe',i:3,': ')
                else write(sortie,'     Group ',i:3,': ');
      count:=0;
      e:=0;
      cardiint:=cardinalite[i];
      cardireel:=cardiint;
      for j:=1 to nobj do if posgroupes[j]=i then
            begin
            if cardiint > 1 then
                  begin
                  for k:=1 to nbvar do
                        e:=e+sqr(donnees[j,k]-moyenne[i,k]/cardireel);
                  if count mod 13 = 0 then
                        if count > 1 then write(sortie,' ':16);
                  end;
            write(sortie,j:4);
            count:=count+1;
            if count mod 13 = 0 then writeln(sortie)
            end;
      if count mod 13 <> 0 then writeln(sortie);
      writeln(sortie,'             E = ',e:13:5);
      sume:=sume+e;
      end;
writeln; if lang=1 then writeln(sortie,'     Somme E = ',sume:15:5)
                   else writeln(sortie,'     Sum E   = ',sume:15:5);
end;      (* resultats *)

procedure entre(i,j:integer);
begin
posgroupes[i]:=j;
ajoute(i,groupes[j]);
end;

procedure lisgroupes(initial:boolean);
var i,j:integer;
begin
if initial then
      begin
      reset(fichclas);
      end;
for i:=1 to nbgroupes do
      begin
      repeat
            read(fichclas,j);
            if j > 0 then
                  begin
                  if (j  <= nobj) then
                        begin
                        if not(membre(j,groupes[i])) then entre(j,i)
                        else erreur(4,j);
                        end
                  else erreur(2,0);
                  end;
      until j <=0;
      end;
for i:=1 to nobj do if posgroupes[i] <= 0 then
      begin
      if lang=1 then
      writeln(' A quel groupe appartient l'' objet ',i:1,' ? ')
      else writeln(' To what group does object no. ',i:1,' pertain ? ');
      valide(1,nbgroupes,j);
      entre(i,j)
      end;
end;      (* lisgroupes *)

procedure optiongr(initial:boolean);
var i,j:integer;
    termine:boolean;
begin
for i:=1 to nbgroupes do nullvec(groupes[i]);
case option of
      0:repeat
        if initial then geninit;
        for i:=1 to nbgroupes do cardinalite[i]:=0;
        for i:=1 to nobj do
            begin
            j:=trunc(ggubfs(dseed)*nbgroupes)+1;
            entre(i,j);
            cardinalite[j]:=cardinalite[j]+1;
            end;
        termine:=true;
        for i:=1 to nbgroupes do if cardinalite[i] <2 then termine:=false;
        until termine;
      1:lisgroupes(initial);
      2:for i:=1 to nobj do
            begin
            j:=i mod nbgroupes +1;
            entre(i,j);
            end;
      end; (* case *)
end;      (* optiongr *)

procedure lecture;
var i,j,k,classe:integer;
    x,sumx,sumx2,nobjm1:real;
    action,stop:boolean;
begin
rewrite(sortie);
writeln(sortie);
if lang=1 then begin
writeln(sortie,' P R O G R A M M E   K - M E A N S  avec contraintes');
writeln(sortie);
writeln(sortie,' Auteur: Alain Vaudor');
writeln(sortie,' Departement de Sciences biologiques');
writeln(sortie,' Universite de Montreal');
writeln(sortie,' C. P. 6128');
writeln(sortie,' Montreal, Quebec, H3C 3J7');
writeln(sortie);
writeln(sortie);
writeln(' P R O G R A M M E   K - M E A N S  avec contraintes');
writeln;
writeln(' Auteur: Alain Vaudor');
end else begin
writeln(sortie,' P R O G R A M   K - M E A N S  with constraints');
writeln(sortie);
writeln(sortie,' Author: Alain Vaudor');
writeln(sortie,' Departement de Sciences biologiques');
writeln(sortie,' Universite de Montreal');
writeln(sortie,' C. P. 6128');
writeln(sortie,' Montreal, Quebec, H3C 3J7');
writeln(sortie);
writeln(sortie);
writeln(' P R O G R A M   K - M E A N S  with constraints');
writeln;
writeln(' Author: Alain Vaudor');
end;
writeln;
if lang=1 then begin
writeln(' Nombre d''objets   ');valide(2,maxobj,nobj);
writeln(' Nombre de variables');valide(1,maxvar,nbvar);
writeln(' Nombre de groupes  ');valide(2,maxgrp,nbgroupes);
end else begin
writeln(' Number of objects  ');valide(2,maxobj,nobj);
writeln(' Number of variables');valide(1,maxvar,nbvar);
writeln(' Number of groups   ');valide(2,maxgrp,nbgroupes);
end;
for i:=1 to nobj do posgroupes[i]:=0;
nbmots:=nobj div nbbits;
option:=0;
if lang=1 then begin
writeln(' Type de groupement: ');
writeln('   0: Groupement sans contrainte');
writeln('   1: Groupement avec contrainte de contiguite en 1 dimension');
writeln('   2: Groupement avec contraintes generales',
            ' (fichier de liens obligatoire)');
end else begin
writeln(' Type of clustering: ');
writeln('   0: Clustering without constraint');
writeln('   1: Clustering with contiguity constraint in 1 dimension');
writeln('   2: Clustering with general contiguity constraints',
            ' (LINK file requested)');
end;
valide(0,2,typedegroupement);
for i:=1 to nobj do nullvec(voisins[i]);
case typedegroupement of
  0:begin
    for i:=1 to nobj do ajoute(i,voisins[1]);
    for i:=2 to nobj do voisins[i]:=voisins[1];
    for i:=1 to nobj do enleve(i,voisins[i]);
    if lang=1 then begin
    writeln(' Options pour les configurations initiales: ');
    writeln('    0: Au hasard (methode de Stony Brook)');
    writeln('    1: Votre fichier de configuration(s) initiale(s)');
    writeln('    2: Assignation systematique aux '
            ,nbgroupes:1,' groupes (methode MODULO de Spath)');
    end else begin
    writeln(' Options for the initial configurations: ');
    writeln('    0: At random (Stony Brook method)');
    writeln('    1: Your own file of initial configuration(s)');
    writeln('    2: Systematic assignment to the '
            ,nbgroupes:1,' groups (the MODULUS method of Spath)');
    end;
    valide(0,2,option);
    end;
  1:begin
    for i:=1 to nbgroupes do nullvec(groupes[i]);
    for i:= 2 to nobj do
                begin
            ajoute(i-1,voisins[i]);
            ajoute(i,voisins[i-1]);
            end;
      for i:=1 to nobj-1 do
            begin
            ajoute(i+1,voisins[i]);
            ajoute(i,voisins[i+1]);
            end;
      repeat
            if lang=1 then writeln
     (' Nombre d''objets dans chacun des groupes successifs:')
            else writeln
     (' Number of objects in each of the successive groups:');
            classe:=0;
            for i:=1 to  nbgroupes do
                  begin
                  read(j);
                  classe:=classe+j;
                  cardinalite[i]:=j;
                  end;
            readln;
            if classe <> nobj then
     begin
     if lang=1 then
     writeln(' Erreur: Somme des gr. differente du nb. d'' objets')
            else writeln
     (' Error: Sum of the groups differs from the n. of objects');
     end;
      until classe=nobj;
      classe:=0;
      for i:=1 to nbgroupes do
            begin
            k:=cardinalite[i];
            for j:=classe+1 to classe+k do entre(j,i);
            classe:=classe+k;
            end;
      end;
  2:begin
    reset(lien);
    while not(eof(lien)) do
      begin
(*    while not((lien@ in['0'..'9'])or(eof(lien))) do get(lien); *)
(* On IBM mainframes, the line above is replaced by next 9 lines *)
      stop:=false;
      while not (stop) do
            begin
     if eof(lien) then stop:=true;
     if (not(stop)) then
          begin
   if (not(lien@ in ['0'..'9'])) then get(lien)
   else stop:=true;
   end;
            end;
      if not(eof(lien)) then
            begin
            read(lien,j,k);
            ajoute(k,voisins[j]);
            ajoute(j,voisins[k]);
            end;
      end;
    if lang=1 then begin
    writeln(' Options:');
    writeln('   1: Au hasard (methode de Stony Brook)');
    writeln('   2: Votre fichier de configuration(s) initiale(s)');
    end else begin
    writeln(' Options:');
    writeln('   1: At random (Stony Brook method)');
    writeln('   2: Your own file of initial configuration(s)');
    end;
    valide(1,2,i);
    fichierdegroupes:=i=2;
    end;
  end;  (* case *)
if (typedegroupement in [0,2]) and (option <> 2) then
      begin
      if lang=1 then writeln(' Nombre d''essais ? ')
                else writeln(' Number of iterations ? ');
      readln(nbessais);
      end
else nbessais:=1;
writeln(' Options:');
if lang=1 then begin
writeln('   1: Imprimer tous les resultats intermediaires');
writeln('   2: Configurations initiale et finale seulement');
end else begin
writeln('   1: Print all intermediate results');
writeln('   2: Print results only for initial and final configurations');
end;
valide(1,2,i);
intermediaire:=i=1;
reset(entree);
for i:=1 to nobj do
      for j:=1 to nbvar do read(entree,donnees[i,j]);
if nbvar > 1 then
      begin
      writeln(' Options:');
      if lang=1 then begin
      writeln('   0: Pas de transformation des donnees');
      writeln('   1: Transformation en variables centrees reduites');
      end else begin
      writeln('   0: No data transformation');
      writeln('   1: Transformation to standardized variables');
      end;
      valide(0,1,classe);
      if classe = 1 then
         begin
         nobjm1:=nobj-1;
         for j:=1 to nbvar do
            begin
            sumx:=0;sumx2:=0;
            for i:=1 to nobj do
                  begin
                  x:=donnees[i,j];
                  sumx:=sumx+x;
                  sumx2:=sumx2+sqr(x);
                  end;
            sumx:=sumx/nobj;
            sumx2:=sqrt((sumx2-sqr(sumx)*nobj)/nobjm1);
            if sumx2 < epsilon then erreur(3,j);
            for i:=1 to nobj do donnees[i,j]:=(donnees[i,j]-sumx)/sumx2;
            end;
         end;
      end;
(* Impression des cartes. Esther Beneish, juillet 1995 *)
IF LANG=1 THEN WRITELN(' Desirez-vous les cartes? (O ou N)  ')
          ELSE WRITELN(' Do you want the maps printed? (Y or N) ');
WHILE input@ =' ' DO get(INPUT);                                                
GRAPHE:=input@ IN ['0','O','o','Y','y'];                                         
READLN;                                                                         
IF  GRAPHE THEN
        BEGIN                                                                   
        IF LANG=1 THEN WRITELN                                                  
        (' Largeur des cartes (en caracteres, sans compter le cadre): ')        
        ELSE WRITELN                                                            
        (' Width of the maps (in characters, without the frame): ');
        READLN(LARGEUR);                                                  
        if lang=1 then writeln
        (' Voulez-vous l''axe des x a l''horizontale? (O ou N) ')
        else writeln
        (' Do you want the x axis to be horizontal? (Y or N) ');
        WHILE input@ =' ' DO get(INPUT);
        change:=input@ IN ['0','O','o','Y','y'];                                         
        READLN;                                                                         
        end;
end;      (* lecture *)

procedure prepare(initial:boolean);
var   i,j,k,ancien,reste:integer;
      action:boolean;
      dejafait:vecteurbit;
      x:real;

      procedure croissance(var dejafait:vecteurbit;i,cardgri:integer);
      var j,compte:integer;
          temp,copy1,copy2:vecteurbit;
      begin
      trouve[i]:=false;
      copy1:=groupes[i];
      creefrontiere(copy2,copy1);
      difference(copy2,dejafait);
      if not(vide(copy2)) then
          begin
          for j:=1 to nbgroupes do if j <> i then
                begin
            temp:=copy2;
            union(temp,dejafait);
            if inclusegal(groupes[j],temp) then
                  difference(copy2,groupes[j]);
            end;
          end;
      if not(vide(copy2)) then
            begin
            compte:=cardvect(copy2);
            j:=0;
            while compte > cardgri do
                  begin
                  j:=j+1;
                  if membre(j,copy2) then
                        begin
                        compte:=compte-1;
                        enleve(j,copy2);
                        end;
                  end;
            if not(vide(copy2)) then
                  begin
                  trouve[i]:=true;
                  action:=true;
                  union(dejafait,copy2);
                  union(groupes[i],copy2);
                  cardinalite[i]:=cardgri-compte;
                  end;
            end;
      end;

begin
case typedegroupement of
  0:optiongr(initial);
  1:begin
    end;
  2:      begin
      for i:=1 to nbgroupes do nullvec(groupes[i]);
      if fichierdegroupes then
            begin
            lisgroupes(initial);
            if not(connexe) then erreur(5,0);
            end
      else
         begin
         nullvec(dejafait);
         for i:=1 to nobj do posgroupes[i]:=0;
         for i:=1 to 5 do x:=ggubfs(dseed);
         if initial then geninit;
         ancien:=0;reste:=0;
         for i:=1 to nbgroupes do
            begin
            reste:=(nobj-ancien) div(nbgroupes-i+1);
            cardinalite[i]:=reste-1;
            ancien:=reste+ancien;
            repeat
                  j:=trunc(ggubfs(dseed)*nobj)+1;
            until not(membre(j,dejafait));
            ajoute(j,dejafait);
            ajoute(j,groupes[i]);
            trouve[i]:=true;
            end;
         repeat
               action:=false;
            for i:=1 to nbgroupes do
               if trouve[i] then croissance(dejafait,i,cardinalite[i]);
         until not(action);
         for i:=1 to nbgroupes do
            for j:=1 to nobj do
               if membre(j,groupes[i]) then posgroupes[j]:=i;
         repeat
               action:=false;
               for i:=1 to nobj do if posgroupes[i]=0 then
               begin
               action:=true;
               ancien:=0;
               for j:=1 to nobj do if membre(j,voisins[i]) then
                  begin
                  k:=posgroupes[j];
                  if k > 0 then ancien:=k;
                  end;
               if ancien > 0 then entre(i,ancien);
               end;
         until not(action);
         end;
    end;
  end;  (* case *)
for i:=1 to nbgroupes do
      begin
      for j:=1 to nbvar do moyenne[i,j]:=0;
      cardinalite[i]:=0;
      end;
for i:=1 to nobj do
      begin
      k:=posgroupes[i];
      cardinalite[k]:=cardinalite[k]+1;
      for j:=1 to nbvar do moyenne[k,j]:=moyenne[k,j]+donnees[i,j];
      end;
resultats(0);
for i:=1 to nbgroupes do creefrontiere(frontiere[i],groupes[i]);
end;

function esp(groupe,objet,signe:integer):real;
var i:integer;
    sum,cardi,cardsigne:real;
begin
i:=cardinalite[groupe];
if i+signe >= 1 then
      begin
      cardi:=i;
      cardsigne:=cardi+signe;
      sum:=0;
      for i:=1 to nbvar do
            sum:=sum+sqr(moyenne[groupe,i]/cardi-donnees[objet,i]);
      esp:=sum*cardi/cardsigne;
      end
else esp:=0;
end;

procedure deplace;
var i,j,gr,ptr,prof:integer;
    esperance,minesp,x:real;
    deplacement:boolean;

      function connecte(obj,gr:integer):boolean;
      var i:integer;
          p1,p2,p3:vecteurbit;
          fini:boolean;
      begin
      if cardinalite[gr]=1 then connecte:=true
      else
         begin
         i:=0;
         p1:=groupes[gr];
         enleve(obj,p1);
         repeat i:=i+1;until membre(i,p1);
         nullvec(p3);
         ajoute(i,p3);
         repeat
            nullvec(p2);
            for i:=1 to nobj do if membre(i,p3) then  union(p2,voisins[i]);
            intersection(p2,p2,p1);
            difference(p2,p3);
            union(p3,p2);
            fini:=egal(p1,p3);
         until (fini)or(vide(p2));
         connecte:=fini;
         end;
      end;

begin
prof:=0;
repeat
      deplacement:=false;
      for i:=1 to nobj do
            begin
            gr:=posgroupes[i];
            minesp:=esp(gr,i,-1);
            ptr:=0;
            if connecte(i,gr) then 
	    for j:=1 to nbgroupes do if membre(i,frontiere[j]) then
                  begin
                  esperance:=esp(j,i,1);
                  if minesp > esperance then
                        begin
                        minesp:=esperance;
                        ptr:=j;
                        end;
                  end;
            if ptr > 0 then
                  begin
                  deplacement:=true;
                  cardinalite[ gr]:=cardinalite[ gr]-1;
                  cardinalite[ptr]:=cardinalite[ptr]+1;
                  enleve(i,groupes[ gr]);
                  entre(i,ptr);
                  creefrontiere(frontiere[gr],groupes[gr]);
                  union(frontiere[ptr],voisins[i]);
                  difference(frontiere[ptr],groupes[ptr]);
                  for j:=1 to nbvar do
                        begin
                        x:=donnees[i,j];
                        moyenne[ gr,j]:=moyenne[ gr,j]-x;
                        moyenne[ptr,j]:=moyenne[ptr,j]+x;
                        end;
                  end;
            end;
      prof:=prof+1;
      if (deplacement) and (intermediaire) then resultats(prof);
      until not(deplacement);
      if not(intermediaire) then resultats(-(prof-1));
      if(prof=1) then begin
          writeln(sortie);
          if lang=1 then writeln(sortie,
              '   La solution initiale n''a pas pu etre amelioree.')
          else writeln(sortie,
              '   The initial solution could not be improved.');
          end;
end;      (* deplace *)

procedure dessine;
VAR I,J,K,NBCAR1,NN,N1,A,B:integer;
    MAXX,MINX,MAXY,MINY,X,Y:double;                                              
    facteurv,facteurh:double;
    coordint:coordtype;
    adresse:memtype;
    alfabet:array[1..30] of char;
    matcar:array[0..132] of char;
 

PROCEDURE SHELL(VAR A: COORDTYPE;N: integer);
LABEL 1;                                                                        
VAR I,J,K,L,M,O,W: integer;
BEGIN                                                                           
M:=(2*N-1) DIV 2;                                                               
WHILE M<>0 DO                                                                   
        BEGIN                                                                   
        K:=N-M;                                                                 
        FOR J:=1 TO K DO                                                        
                   BEGIN
                   I:=J;                                                        
                   WHILE I>=1 DO                                                
                                BEGIN                                           
                                IF A[2,I+M] <= A[2,I] THEN GOTO 1;              
                                O:=I+M;                                         
                                W:=A[1,I];A[1,I]:=A[1,O];A[1,O]:=W;             
                                W:=A[2,I];A[2,I]:=A[2,O];A[2,O]:=W;             
                                W:=ADRESSE[I];ADRESSE[I]:=ADRESSE[O];           
                                ADRESSE[O]:=W;                             
                                I:=I-M;                                         
                                END;                                            
1:                   END;                                                       
        M:=M  DIV 2;                                                            
        END;                                                                    
END;        (* FIN DE SHELL *)                                                  

procedure init;
var i:integer;

begin
        change:=false;
        RESET(coordxy);
        READ(coordxy,X,Y);
        MAXX:=X;MINX:=X;                                                        
        MINY:=Y;MAXY:=Y;                                                        
        FOR I:=2 TO nobj DO
                BEGIN                                                           
                READ(coordxy,X,Y);
                IF X > MAXX THEN MAXX:=X;                                       
                IF X < MINX THEN MINX:=X;                                       
                IF Y > MAXY THEN MAXY:=Y;                                       
                IF Y < MINY THEN MINY:=Y;                                       
                END;                                                            
        MAXX:=MAXX-MINX;                                                        
        MAXY:=MAXY-MINY;                                                        
        IF  change THEN
                BEGIN                                                          
                X:=MAXX;MAXX:=MAXY;MAXY:=X;
                X:=MINX;MINX:=MINY;MINY:=X;                                     
                END;                                            
        FACTEURH:=LARGEUR/(MAXX+1);                                             
        FACTEURV:=FACTEURH/FACTEUR;                                             
        LARGEUR:=LARGEUR+2;                                                     
        RESET(coordxy);
        FOR I:=1 TO nobj DO                                                  
                BEGIN                                                          
                READ(coordxy,X,Y);                                      
                COORDINT[1,I]:=ROUND((X-MINX)*FACTEURH);                        
                COORDINT[2,I]:=ROUND((Y-MINY)*FACTEURV);                        
                ADRESSE[I]:=I;                                            
                END;                                                           
        SHELL(COORDINT,Nobj);                                                
        COORDINT[2,nobj+1]:=-1;
        for i:=1 to 25 do alfabet[i]:=chr(97+i-1);
        for i:=26 to 30 do alfabet[i]:=chr(i-25);
                                            
end;

PROCEDURE ECRITURE;
VAR I,J,MAXPT,PT:integer;

        PROCEDURE ECRITLIGNE;
        VAR I:integer;
        BEGIN                                                                   
        WRITE(SORTIE,' !');         
        FOR I:=2 TO LARGEUR-1 DO WRITE(SORTIE,' ');                             
        WRITELN(SORTIE,'!');                                                    
        END;                                                                    
                                                                                
        PROCEDURE ECRITTRAIT;                                                   
        VAR I:integer;
        BEGIN                                                                   
        WRITE(SORTIE,' ');                                                      
        FOR I:=1 TO LARGEUR DO WRITE(SORTIE,'-');                               
        WRITELN(SORTIE);                                                        
        END;                                                                    

        begin
        IF GRAPHE THEN
                BEGIN                                                         
                PT:=1;                                                          
                MAXPT:=COORDINT[2,1];                                           
                ECRITTRAIT;                                                     
                WHILE MAXPT >= 0 DO                                             
                        BEGIN                                                   
                        FOR I:=0 TO LARGEUR-3 DO MATCAR[I]:=' ';                
                        WHILE COORDINT[2,PT]=MAXPT DO                           
                                BEGIN                                           
                                I:=posgroupes[ADRESSE[PT]];
                                IF I <=maxgrp THEN                          
                                     MATCAR[COORDINT[1,PT]]:=ALFABET[I];        
                                PT:=PT+1;                                       
                                END;                                            
                        WRITE(SORTIE,' !');                                     
                        FOR I:=0 TO LARGEUR-3 DO WRITE(SORTIE,MATCAR[I]);                      
                        WRITELN(SORTIE,'!');                                    
                        IF MAXPT > 0 THEN                                       
                                BEGIN                                           
                                FOR I:=MAXPT-2 DOWNTO COORDINT[2,PT] DO         
                                    ECRITLIGNE;                                 
                                MAXPT:=COORDINT[2,PT];                          
                                END                                             
                        ELSE MAXPT:=-1;                                         
                        END;                                                    
                ECRITTRAIT;
                WRITELN(SORTIE);                                                
                END;                                                            
        WRITELN(SORTIE);                                                        
        END;                                                                   

begin  {* dessine *}
init;
ecriture;
end;   {* dessine *}

begin      (* prog princ *)

dseed:=5.98763;   
lecture;
for noboucle:=1 to nbessais do
      begin
      writeln(sortie);writeln(sortie);
      if nbessais > 1 then
           begin
           writeln;
           if lang=1 then writeln(sortie,' Essai no ',noboucle:1)
                     else writeln(sortie,' Iteration no. ',noboucle:1);
           end;
      prepare(noboucle = 1);
      deplace;
      end;
if graphe then dessine;
if lang=1 then writeln(' Fin du programme. ')
          else writeln(' End of the program. ');
end.
